import numpy as np
import pandas as pd
import scipy
from scipy.stats import norm, uniform
import unittest
import networkx as nx

from context import grama as gr
from context import models

## FD stepsize
h = 1e-8

## Core function tests
##################################################
class TestModel(unittest.TestCase):
    """Test implementation of model
    """

    def setUp(self):
        # Default model
        self.df_wrong = pd.DataFrame(data={"z": [0.0, 1.0]})

        # 2D identity model with permuted df inputs
        domain_2d = gr.Domain(bounds={"x0": [-1.0, +1.0], "x1": [0.0, 1.0]})
        marginals = {}
        marginals["x0"] = gr.MarginalNamed(
            d_name="uniform", d_param={"loc": -1, "scale": 2}
        )
        marginals["x1"] = gr.MarginalNamed(
            sign=-1, d_name="uniform", d_param={"loc": 0, "scale": 1},
        )

        self.model_2d = gr.Model(
            functions=[
                gr.Function(
                    lambda x0, x1: [x0, x1], ["x0", "x1"], ["y0", "y1"], "test", 0
                ),
            ],
            domain=domain_2d,
            density=gr.Density(marginals=marginals),
        )
        self.df_2d = pd.DataFrame(data={"x1": [0.0], "x0": [+1.0]})
        self.res_2d = self.model_2d.evaluate_df(self.df_2d)

        self.df_median_in = pd.DataFrame({"x0": [0.5], "x1": [0.5]})
        self.df_median_out = pd.DataFrame({"x0": [0.0], "x1": [0.5]})

        self.model_3d = gr.Model(
            functions=[
                gr.Function(
                    lambda x, y, z: x + y + z, ["x", "y", "z"], ["f"], "test", 0
                )
            ],
            density=gr.Density(marginals=marginals),
        )

        ## Timing check
        self.model_slow = gr.Model(
            functions=[
                gr.Function(lambda x: x, ["x0"], ["y0"], "f0", 1),
                gr.Function(lambda x: x, ["x0"], ["y1"], "f1", 1),
            ]
        )

    def test_prints(self):
        ## Invoke printpretty
        self.model_3d.printpretty()

    def test_timings(self):
        ## Default is zero
        self.assertTrue(self.model_2d.runtime(1) == 0)

        ## Estimation accounts for both functions
        self.assertTrue(np.allclose(self.model_slow.runtime(1), 2))

        ## Slow function returns string message
        msg = self.model_slow.runtime_message(pd.DataFrame({"x0": [0]}))
        self.assertTrue(isinstance(msg, str))

        ## Units switch to minutes
        msg = self.model_slow.runtime_message(pd.DataFrame({"x0": [0] * 60}))
        self.assertTrue(gr.str_detect(msg, "min"))

        ## Units switch to hours
        msg = self.model_slow.runtime_message(pd.DataFrame({"x0": [0] * 3600}))
        self.assertTrue(gr.str_detect(msg, "hr"))

    ## Basic functionality with default arguments

    def test_catch_input_mismatch(self):
        """Checks that proper exception is thrown if evaluate(df) passed a
        DataFrame without the proper columns.
        """
        self.assertRaises(ValueError, self.model_2d.evaluate_df, self.df_wrong)

    def test_var_outer(self):
        ## Test pass-throughs
        df_test = pd.DataFrame(dict(x0=[0]))
        md_no_rand = gr.Model() >> gr.cp_function(fun=lambda x: x, var=1, out=1)
        md_no_rand.var_outer(pd.DataFrame(), df_det="nom")

        md_no_det = md_no_rand >> gr.cp_marginals(
            x0={"dist": "uniform", "loc": 0, "scale": 1}
        )
        md_no_det.var_outer(df_test, df_det="nom")

        ## Test assertions
        with self.assertRaises(ValueError):
            self.model_3d.var_outer(self.df_2d, df_det=self.df_2d)

    def test_drop_out(self):
        """Checks that output column names are properly dropped"""
        md = gr.Model() >> gr.cp_function(lambda x0: x0 + 1, var=1, out=1)
        df_in = gr.df_make(x0=[0, 1, 2], y0=[0, 1, 2])
        df_true = gr.df_make(x0=[0, 1, 2], y0=[1, 2, 3])

        df_res = md >> gr.ev_df(df=df_in)

        self.assertTrue(gr.df_equal(df_res, df_true, close=True))

    ## Test re-ordering issues

    def test_2d_output_names(self):
        """Checks that proper output names are assigned to resulting DataFrame
        """
        self.assertEqual(
            set(self.model_2d.evaluate_df(self.df_2d).columns), set(self.model_2d.out)
        )

    def test_quantile(self):
        """Checks that model.sample_quantile() evaluates correctly.
        """
        df_res = self.model_2d.density.pr2sample(self.df_median_in)

        self.assertTrue(gr.df_equal(df_res, self.df_median_out))

    def test_empty_functions(self):
        md = gr.Model() >> gr.cp_bounds(x=[-1, +1])
        with self.assertRaises(ValueError):
            gr.eval_nominal(md)

    def test_nominal(self):
        """Checks the implementation of nominal values"""
        md = gr.Model() >> gr.cp_bounds(
            x0=[-1, +1], x1=[0.1, np.Inf], x2=[-np.Inf, -0.1],
        )
        df_true = gr.df_make(x0=0.0, x1=+0.1, x2=-0.1)
        df_res = gr.eval_nominal(md, df_det="nom", skip=True)

        self.assertTrue(gr.df_equal(df_res, df_true))

    ## Test sample transforms
    def test_transforms(self):
        ## Setup
        df_corr = pd.DataFrame(dict(var1=["x"], var2=["y"], corr=[0.5]))
        Sigma_h = np.linalg.cholesky(np.array([[1.0, 0.5], [0.5, 1.0]]))

        md = (
            gr.Model()
            >> gr.cp_marginals(
                x=dict(dist="norm", loc=0, scale=1), y=dict(dist="norm", loc=0, scale=1)
            )
            >> gr.cp_copula_gaussian(df_corr=df_corr)
        )

        ## Copula and marginals have same var_rand order
        self.assertTrue(list(md.density.marginals) == md.density.copula.var_rand)

        ## Transforms invariant
        z = np.array([0, 0])
        x = md.z2x(z)
        zp = md.x2z(x)

        self.assertTrue(np.all(z == zp))

        df_z = gr.df_make(x=0.0, y=0.0)
        df_x = md.norm2rand(df_z)
        df_zp = md.rand2norm(df_x)

        self.assertTrue(gr.df_equal(df_z, df_zp))

        ## Jacobian accurate
        dxdz_fd = np.zeros((2, 2))
        dxdz_fd[0, :] = (md.z2x(z + np.array([h, 0])) - md.z2x(z)) / h
        dxdz_fd[1, :] = (md.z2x(z + np.array([0, h])) - md.z2x(z)) / h
        dxdz_p = md.dxdz(z)

        self.assertTrue(np.allclose(dxdz_fd, dxdz_p))

    ## Test DAG construction

    def test_dag(self):
        md = (
            gr.Model("model")
            >> gr.cp_function(lambda x: x, var=1, out=1)
            >> gr.cp_function(lambda x0, y0: x0 + y0, var=["x0", "y0"], out=1)
        )

        G_true = nx.DiGraph()
        G_true.add_edge("(var)", "f0", label="{}".format({"x0"}))
        G_true.add_edge("f0", "(out)", label="{}".format({"y0"}))
        G_true.add_edge("(var)", "f1", label="{}".format({"x0"}))
        G_true.add_edge("f0", "f1", label="{}".format({"y0"}))
        G_true.add_edge("f1", "(out)", label="{}".format({"y1"}))
        nx.set_node_attributes(G_true, "model", "parent")

        self.assertTrue(
            nx.is_isomorphic(
                md.make_dag(),
                G_true,
                node_match=lambda u, v: u == v,
                edge_match=lambda u, v: u == v,
            )
        )


class TestEvalDf(unittest.TestCase):
    """Test implementation of eval_df()
    """

    def setUp(self):
        self.model = models.make_test()
        self.df = gr.df_make(x=1)


    def test_catch_wrong_type(self):
        """Checks that eval_df() raises when wrong input md or df arg is given.

        checks: None, tuple, str, list
        """
        tests = [None, (1,2), 2, "a", [1, 8]]
        # wrong df arg
        for wrong_type in tests:
            self.assertRaises(TypeError, gr.eval_df, self.model, wrong_type)

        # wrong md arg
        for wrong_type in tests:
            self.assertRaises(TypeError, gr.eval_df, wrong_type, self.df)


# --------------------------------------------------
class TestDomain(unittest.TestCase):
    def setUp(self):
        self.domain = gr.Domain(bounds={"x": (0, 1)})

    def test_blank(self):
        ## Test blank domain valid
        gr.Domain()

        ## Invoke summary
        self.domain.bound_summary("x")

        ## Invoke summary;
        self.assertTrue(self.domain.bound_summary("y").find("unbounded") > -1)


# --------------------------------------------------
class TestDensity(unittest.TestCase):
    def setUp(self):
        self.density = gr.Density(
            marginals=dict(
                x=gr.MarginalNamed(d_name="uniform", d_param={"loc": -1, "scale": 2}),
                y=gr.MarginalNamed(d_name="uniform", d_param={"loc": -1, "scale": 2}),
            ),
            copula=gr.CopulaGaussian(
                ["x", "y"], pd.DataFrame(dict(var1=["x"], var2=["y"], corr=[0.5]))
            ),
        )

        self.density_ind = gr.Density(
            marginals=dict(
                x=gr.MarginalNamed(d_name="uniform", d_param={"loc": -1, "scale": 2}),
                y=gr.MarginalNamed(d_name="norm", d_param={"loc": 0, "scale": 1}),
            ),
            copula=gr.CopulaIndependence(["x", "y"]),
        )
        self.density_gauss = gr.Density(
            marginals=dict(
                x=gr.MarginalNamed(d_name="uniform", d_param={"loc": -1, "scale": 2}),
                y=gr.MarginalNamed(d_name="norm", d_param={"loc": 0, "scale": 1}),
            ),
            copula=gr.CopulaGaussian(
                ["x", "y"], pd.DataFrame(dict(var1=["x"], var2=["y"], corr=[0.5]))
            ),
        )

    def test_copula_warning(self):
        md = gr.Model()

        with self.assertRaises(ValueError):
            md.density.sample()

    def test_CopulaIndependence(self):
        copula = gr.CopulaIndependence(var_rand=["x", "y"])
        df_res = copula.sample(seed=101)

        self.assertTrue(set(df_res.columns) == set(["x", "y"]))

        ## Transforms invariant
        z = np.array([0, 0])
        u = copula.z2u(z)
        zp = copula.u2z(u)

        self.assertTrue(np.all(z == zp))

        ## Jacobian accurate
        dudz_fd = np.zeros((2, 2))
        dudz_fd[0, :] = (copula.z2u(z + np.array([h, 0])) - copula.z2u(z)) / h
        dudz_fd[1, :] = (copula.z2u(z + np.array([0, h])) - copula.z2u(z)) / h
        dudz_p = copula.dudz(z)

        self.assertTrue(np.allclose(dudz_fd, dudz_p))

    def test_CopulaGaussian(self):
        df_corr = pd.DataFrame(dict(var1=["x"], var2=["y"], corr=[0.5]))
        Sigma_h = np.linalg.cholesky(np.array([[1.0, 0.5], [0.5, 1.0]]))
        copula = gr.CopulaGaussian(["x", "y"], df_corr=df_corr)
        df_res = copula.sample(seed=101)

        self.assertTrue(np.isclose(copula.Sigma_h, Sigma_h).all)
        self.assertTrue(set(df_res.columns) == set(["x", "y"]))

        ## Test raises
        df_corr_invalid = pd.DataFrame(
            dict(var1=["x", "x"], var2=["y", "z"], corr=[0, 0])
        )

        with self.assertRaises(ValueError):
            gr.CopulaGaussian(["x", "y"], df_corr=df_corr_invalid)

        ## Transforms invariant
        z = np.array([0, 0])
        u = copula.z2u(z)
        zp = copula.u2z(u)

        self.assertTrue(np.all(z == zp))

        ## Jacobian accurate
        dudz_fd = np.zeros((2, 2))
        dudz_fd[0, :] = (copula.z2u(z + np.array([h, 0])) - copula.z2u(z)) / h
        dudz_fd[1, :] = (copula.z2u(z + np.array([0, h])) - copula.z2u(z)) / h
        dudz_p = copula.dudz(z)

        self.assertTrue(np.allclose(dudz_fd, dudz_p))

    def test_conversion(self):
        df_pr_true = pd.DataFrame(dict(x=[0.5], y=[0.5]))
        df_sp_true = pd.DataFrame(dict(x=[0.0], y=[0.0]))

        df_pr_res = self.density.sample2pr(df_sp_true)
        df_sp_res = self.density.pr2sample(df_pr_true)

        self.assertTrue(gr.df_equal(df_pr_true, df_pr_res))
        self.assertTrue(gr.df_equal(df_sp_true, df_sp_res))

    def test_sampling(self):
        df_sample = self.density.sample(n=1, seed=101)

        self.assertTrue(set(df_sample.columns) == set(["x", "y"]))

    def test_density(self):
        x = np.array([-0.9, -0.5, 0, +0.5, +0.9])
        y = np.array([-0.9, -0.5, 0, +0.5, +0.9])
        df = pd.DataFrame(dict(x=x, y=y))

        ## Independence copula
        # Exact
        l_true = (
            1
            * scipy.stats.uniform(loc=-1, scale=2).pdf(x)
            * scipy.stats.norm(loc=0, scale=1).pdf(y)
        )
        # Computed
        l_comp = self.density_ind.d(df)

        self.assertTrue(all(l_true == l_comp))

        ## Gaussian copula
        R = np.array([[1.0, 0.5], [0.5, 1.0]])
        R_inv = np.linalg.inv(R)
        det = np.linalg.det(R)
        I = np.eye(2)
        df_u = self.density_gauss.sample2pr(df)

        l_gauss_cop = np.zeros(len(x))
        for i in range(len(x)):
            l_gauss_cop[i] = np.exp(
                -0.5 * np.dot(norm.ppf(df_u.values[i]), np.dot(R_inv - I, norm.ppf(df_u.values[i])))
            ) / np.sqrt(det)
        # Exact
        l_true = (
            l_gauss_cop
            * scipy.stats.uniform(loc=-1, scale=2).pdf(x)
            * scipy.stats.norm(loc=0, scale=1).pdf(y)
        )
        # Computed
        l_comp = self.density_gauss.d(df)

        self.assertTrue(np.allclose(l_true, l_comp, rtol=1e-3, atol=1e-3))

# --------------------------------------------------
class TestFunction(unittest.TestCase):
    def setUp(self):
        self.fcn = gr.Function(lambda x: x, ["x"], ["x"], "test", 0)

        self.fcn_vec = gr.FunctionVectorized(lambda df: df, ["x"], ["x"], "test", 0)

        self.df = pd.DataFrame({"x": [0]})

        self.df_wrong = pd.DataFrame({"z": [0]})

    def test_function(self):
        fcn_copy = self.fcn.copy()

        self.assertTrue(self.fcn.var == fcn_copy.var)
        self.assertTrue(self.fcn.out == fcn_copy.out)
        self.assertTrue(self.fcn.name == fcn_copy.name)

        pd.testing.assert_frame_equal(
            self.df, self.fcn.eval(self.df), check_dtype=False
        )

        with self.assertRaises(ValueError):
            self.fcn.eval(self.df_wrong)

        ## Invoke summary
        self.fcn.summary()

    def test_function_vectorized(self):
        fcn_copy = self.fcn_vec.copy()

        self.assertTrue(self.fcn_vec.var == fcn_copy.var)
        self.assertTrue(self.fcn_vec.out == fcn_copy.out)
        self.assertTrue(self.fcn_vec.name == fcn_copy.name)

        pd.testing.assert_frame_equal(
            self.df, self.fcn_vec.eval(self.df), check_dtype=False
        )

    def test_function_model(self):
        md_base = gr.Model() >> gr.cp_function(
            fun=lambda x: x, var=1, out=1, name="name", runtime=1
        )

        ## Base constructor
        func = gr.FunctionModel(md_base)

        self.assertTrue(md_base.var == func.var)
        self.assertTrue(md_base.out == func.out)
        self.assertTrue(md_base.name == func.name)
        self.assertTrue(md_base.runtime(1) == func.runtime)

        ## Test copy
        func_copy = func.copy()

        self.assertTrue(func_copy.var == func.var)
        self.assertTrue(func_copy.out == func.out)
        self.assertTrue(func_copy.name == func.name)
        self.assertTrue(func_copy.runtime == func.runtime)


## Run tests
if __name__ == "__main__":
    unittest.main()



from __future__ import print_function, division

import numpy as np
from .cdist import cdist

from .vertex import voronoi_vertex_sample, voronoi_vertex

def fill_distance(domain, Xhat, L = None):
	r""" Compute the exact fill distance by constructing the bounded Voronoi vertices


	Parameters
	----------
	domain: Domain
		Domain on which to compute the dispersion
	Xhat: array-like (?, m)
		Existing samples on the domain
	L: array-like (?, m) optional
		Matrix defining the distance metric on the domain
	
	Returns
	-------
	d: float
		Fill distance 
	"""

	V = voronoi_vertex(domain, Xhat, L = L)
	D = cdist(Xhat, V, L = L)
	return np.max(np.min(D, axis= 0))	

def fill_distance_estimate(domain, Xhat, L = None, Nsamp = int(1e3), X0 = None ):
	r""" Estimate the fill distance of the points Xhat in the domain

	The *fill distance* (Def. 1.4 of [Wen04]_) or *dispersion* [LC05]_
	is the furthest distance between any point :math:`\mathbf{x} \in \mathcal{D}` 
	and a set of points 
	:math:`\lbrace \widehat{\mathbf{x}}_j \rbrace_{j=1}^m \subset \mathcal{D}`:
		
	.. math::

		\sup_{\mathbf{x} \in \mathcal{D}} \min_{j=1,\ldots,M} \|\mathbf{L}(\mathbf{x} - \widehat{\mathbf{x}}_j)\|_2.

	Similar to :meth:`psdr.seq_maximin_sample`, this uses :meth:`psdr.voronoi_vertex_sample` to find 
	a subset of local maximizers and returns the best of these.

	Parameters
	----------
	domain: Domain
		Domain on which to compute the dispersion
	Xhat: array-like (?, m)
		Existing samples on the domain
	L: array-like (?, m) optional
		Matrix defining the distance metric on the domain
	Nsamp: int, default 1e4
		Number of samples to use for vertex sampling
	X0: array-like (?, m)
		Samples from the domain to use in :meth:`psdr.voronoi_vertex_sample`

	Returns
	-------
	d: float
		Fill distance lower bound

	References
	----------
	..  [Wen04] Scattered Data Approximation. Holger Wendland.
			Cambridge University Press, 2004.
			https://doi.org/10.1017/CBO9780511617539
	.. [LC05] Iteratively Locating Voronoi Vertices for Dispersion Estimation
		Stephen R. Lindemann and Peng Cheng
		Proceedings of the 2005 Interational Conference on Robotics and Automation
	"""
	if L is None:
		L = np.eye(len(domain))
	
	if X0 is None:
		from ..sample import initial_sample
		X0 = initial_sample(domain, L, Nsamp = Nsamp)

	# Since we only care about distance in L, we can terminate early if L is rank-deficient
	# and hence we turn off the randomization
	Xcan = voronoi_vertex_sample(domain, Xhat, X0, L = L, randomize = False)

	# Euclidean distance
	D = cdist(Xcan, Xhat, L)

	d = np.min(D, axis = 1)
	return float(np.max(d))




__all__ = [
    "eval_df",
    "ev_df",
    "eval_linup",
    "ev_linup",
    "eval_nominal",
    "ev_nominal",
    "eval_grad_fd",
    "ev_grad_fd",
    "eval_sample",
    "ev_sample",
    "eval_conservative",
    "ev_conservative",
]

import itertools
from grama import add_pipe, tran_outer, custom_formatwarning, Model
from numbers import Integral
from numpy import diag, dot, ones, eye, tile, atleast_2d, triu_indices, round
from numpy.random import seed as set_seed
from pandas import DataFrame, concat
from toolz import curry
from warnings import formatwarning, catch_warnings, simplefilter

formatwarning = custom_formatwarning

def invariants_eval_model(md, skip=False):
    r"""Helper function to group common model argument invariant checks for eval functions.

    Throws errors for invalid Model inputs.

    Args:
        md (gr.Model): Model to check
        skip (bool): if function is skipping evaluation of function. If True,
            skips model.functions test

    """
    ## Type Checking
    if not isinstance(md, Model):
        if md is None:
            raise TypeError("No input model given")
        elif isinstance(md, tuple):
            raise TypeError("Given model argument is type tuple. Have you " +
            "declared your model with an extra comma after the closing `)`?")
        else:
            raise TypeError("Type gr.Model was expected, a " + str(type(md)) +
            " was passed.")

    ## Value checking
    if not skip and len(md.functions) == 0:
        raise ValueError("Given model has no functions.")
    return

def invariants_eval_df(df, arg_name="df", valid_str=None, acc_none=False):
    r"""Helper function to group common DataFrame argument invariant checks for eval functions.

    Throws errors for invalid DataFrame inputs.

    Args:
        df (DataFrame): DataFrame to test
        arg_name (str): Name of df argument
        valid_str (list(str) or None): Valid string inputs, if any, to
            allow when testing
        acc_none (bool): allow `None` as a valid df input

    Examples:
        invariants_eval_df(df)
        invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom", "det"])
        invariants_eval_df(df_test, arg_name="df_test", acc_none=())

    """
    def aps(string):
        r"""Helper function for valid_args_msg() to put apostrophes around a
        string.

        Args:
            string (str): string to surround

        Returns:
            String: Input string surrounded as 'input'"""
        return "'" + string + "'"

    def valid_args_msg(df_arg, acc_str, valid_str):
        r"""Generates string explaining valid inputs for use in DataFrame
        TypeErrors and ValueErrors

        Args:
            df_arg (str): Name of df argument
            acc_str (bool): Indicates whether strings are accepted or not
            valid_str (None, list(str)): Valid string inputs

        Returns:
            String"""
        msg = df_arg + " must be DataFrame" # general msg for valid args
        if acc_str:
            # add on string options to msg
            if len(valid_str) == 1:
                string_args = " or " + aps(valid_str[0])
            else:
                string_args = ", "  # comma after "must be DataFrame"
                for arg in valid_str:
                    if arg == valid_str[-1]:
                        # last value -> add or
                        string_args += "or " + aps(arg)
                    else:
                        # not last value -> add comma
                        string_args += aps(arg) + ", "  # add comma
            msg += string_args + "."
        else:
            # no valid string inputs, end message
            msg += "."
        return msg

    ## Type Checking & String Input
    acc_str = isinstance(valid_str, list)
    if not isinstance(df, DataFrame):
        if df is None:
            if not acc_none:
                # allow "None" df if None accepted
                raise TypeError("No " + arg_name + " argument given. " +
                    valid_args_msg(arg_name, acc_str, valid_str))
        elif isinstance(df, str) and acc_str:
            # case check for invalid str input
            if df not in valid_str:
                raise ValueError(arg_name + " shortcut string invalid. " +
                    valid_args_msg(arg_name, acc_str, valid_str))
        else:
            raise TypeError(valid_args_msg(arg_name, acc_str, valid_str) +
                " Given argument is type " + str(type(df)) +
                    ". ")

    ## Value checking
    #### TO DO

    return



## Default evaluation function
# --------------------------------------------------
@curry
def eval_df(model, df=None, append=True, verbose=True):
    r"""Evaluate model at given values

    Evaluates a given model at a given dataframe.

    Args:
        model (gr.Model): Model to evaluate
        df (DataFrame): Input dataframe to evaluate
        append (bool): Append results to original dataframe?

    Returns:
        DataFrame: Results of model evaluation

    Examples::

        import grama as gr
        from grama.models import make_test
        md = make_test()
        df = gr.df_make(x0=0, x1=1, x2=2)
        md >> gr.ev_df(df=df)

    """
    ## Perform common invariant tests
    invariants_eval_model(model)
    invariants_eval_df(df)

    out_intersect = set(df.columns).intersection(model.out)
    if (len(out_intersect) > 0) and verbose:
        print(
            "... provided columns intersect model output.\n"
            + "eval_df() is dropping {}".format(out_intersect)
        )

    df_res = model.evaluate_df(df)

    if append:
        df_res = concat(
            [
                df.reset_index(drop=True).drop(model.out, axis=1, errors="ignore"),
                df_res,
            ],
            axis=1,
        )

    return df_res


ev_df = add_pipe(eval_df)

## Linear Uncertainty Propagation
# --------------------------------------------------
@curry
def eval_linup(model, df_base=None, append=True, decomp=False, decimals=2, n=1e4, seed=None):
    r"""Linear uncertainty propagation

    Approximates the variance of output models using a linearization of functions---linear uncertainty propagation. Optionally decomposes the output variance according to additive factors from each input.

    Args:
        model (gr.Model): Model to evaluate
        df_base (DataFrame or None): Base levels for evaluation; use
            "nom" for nominal levels.
        append (bool): Append results to nominal inputs?
        decomp (bool): Decompose the fractional variances according to each input?
        decimals (int): Decimals to report for fractional variances

        n (float): Monte Carlo sample size, for estimating covariance matrix
        seed (int or None): Monte Carlo seed

    Returns:
        DataFrame: Output variances at each deterministic level

    Examples::

        import grama as gr
        from grama.models import make_test
        md = make_test()
        ## Set manual levels for deterministic inputs; nominal levels for random inputs
        md >> gr.ev_linup(df_det=gr.df_make(x2=[0, 1, 2])
        ## Use nominal deterministic levels
        md >> gr.ev_linup(df_base="nom")

    """
    ## Perform common invariant tests
    invariants_eval_model(model, False)
    invariants_eval_df(df_base, arg_name="df_base", valid_str=["nom"])

    if df_base is "nom":
        df_base = eval_nominal(model, df_det="nom")
    else:
        df_base = eval_df(model, df=df_base)

    ## Approximate the covariance matrix
    df_sample = eval_sample(model, n=n, seed=seed, df_det=df_base[model.var_det], skip=True)
    cov = df_sample[model.var_rand].cov()

    ## Approximate the gradient
    df_grad = eval_grad_fd(model, df_base=df_base, var="rand", append=True)

    ## Iterate over all outputs and gradient points
    df_res = DataFrame({})
    for out in model.out:
        for i in range(df_grad.shape[0]):
            # Build gradient column names
            var_names = map(
                lambda s: "D" + out + "_D" + s,
                model.var_rand
            )
            # Extract gradient values
            grad = df_grad.iloc[i][var_names].values.flatten()
            # Approximate variance
            var = dot(grad, dot(cov, grad))

            # Store values
            df_tmp = df_base.iloc[[i]][model.var + model.out].reset_index(drop=True)
            df_tmp["out"] = out
            df_tmp["var"] = var

            # Decompose variance due to each input
            if decomp:
                grad_mat = diag(grad)
                sens_mat = dot(grad_mat, dot(cov, grad_mat)) / var
                U, V = triu_indices(sens_mat.shape[0])
                sens_values = [
                    round(
                        sens_mat[U[j], V[j]] * (1 + (U[j] != V[j])), # Double off-diag
                        decimals=decimals
                    )
                    for j in range(len(U))
                ]
                sens_var = [
                    model.var_rand[U[j]] + "*" + model.var_rand[V[j]]
                    for j in range(len(U))
                ]

                df_sens = DataFrame({
                    "var_frac": sens_values,
                    "var_rand": sens_var
                })
                df_tmp = tran_outer(df_tmp, df_sens)

            df_res = concat(
                (df_res,df_tmp),
                axis=0,
            ).reset_index(drop=True)

    return df_res

ev_linup = add_pipe(eval_linup)


## Nominal evaluation
# --------------------------------------------------
@curry
def eval_nominal(model, df_det=None, append=True, skip=False):
    r"""Evaluate model at nominal values

    Evaluates a given model at a model nominal conditions (median) of random inputs. Optionally set nominal values for the deterministic inputs.

    Args:
        model (gr.Model): Model to evaluate
        df_det (DataFrame or None): Deterministic levels for evaluation; use
            "nom" for nominal deterministic levels. If provided model has no
            deterministic variables (model.n_var_det == 0), then df_det may
            equal None.
        append (bool): Append results to nominal inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Results of nominal model evaluation or unevaluated design

    Examples::

        import grama as gr
        from grama.models import make_test
        md = make_test()
        ## Set manual levels for deterministic inputs; nominal levels for random inputs
        md >> gr.ev_nominal(df_det=gr.df_make(x2=[0, 1, 2])
        ## Use nominal deterministic levels
        md >> gr.ev_nominal(df_det="nom")

    """
    ## Perform common invariant tests
    invariants_eval_model(model, skip)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom"],
        acc_none=(model.n_var_det==0))

    ## Draw from underlying gaussian
    quantiles = ones((1, model.n_var_rand)) * 0.5  # Median

    ## Convert samples to desired marginals
    df_pr = DataFrame(data=quantiles, columns=model.var_rand)
    df_rand = model.density.pr2sample(df_pr)
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)

    if skip:
        ## Evaluation estimate
        runtime_msg = model.runtime_message(df_samp)
        print(runtime_msg)

        return df_samp
    return eval_df(model, df=df_samp, append=append)


ev_nominal = add_pipe(eval_nominal)

## Gradient finite-difference evaluation
# --------------------------------------------------
@curry
def eval_grad_fd(model, h=1e-8, df_base=None, var=None, append=True, skip=False):
    r"""Finite-difference gradient approximation

    Evaluates a given model with a central-difference stencil to approximate the
    gradient.

    Args:
        model (gr.Model): Model to differentiate
        h (numeric): finite difference stepsize,
            single (scalar): or per-input (array)
        df_base (DataFrame): Base-points for gradient calculations
        var (list(str) or string): list of variables to differentiate,
            or flag; "rand" for var_rand, "det" for var_det
        append (bool): Append results to base point inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Gradient approximation or unevaluated design

    @pre (not isinstance(h, collections.Sequence)) |
         (h.shape[0] == df_base.shape[1])

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md = make_cantilever_beam()
        # Select base point(s)
        df_nom = md >> gr.ev_nominal(df_det="nom")
        # Approximate the gradient
        df_grad = md >> gr.ev_grad_fd(df_base=df_nom)

    """
    ## Check invariants
    invariants_eval_model(model, skip)
    invariants_eval_df(df_base, arg_name="df_base")
    if not set(model.var).issubset(set(df_base.columns)):
        raise ValueError("model.var must be subset of df_base.columns")
    if var is None:
        var = model.var
    elif isinstance(var, str):
        if var == "rand":
            var = model.var_rand
        elif var == "det":
            var = model.var_det
        else:
            raise ValueError("var flag not recognized; use 'rand' or 'det'")
    else:
        if not set(var).issubset(set(model.var)):
            raise ValueError("var must be subset of model.var")
    var_fix = list(set(model.var).difference(set(var)))

    ## TODO
    if skip == True:
        raise NotImplementedError("skip not implemented")

    ## Build stencil
    n_var = len(var)
    stencil = eye(n_var) * h
    stepscale = tile(atleast_2d(0.5 / h).T, (1, model.n_out))

    outputs = model.out
    nested_labels = [
        list(map(lambda s_out: "D" + s_out + "_D" + s_var, outputs)) for s_var in var
    ]
    grad_labels = list(itertools.chain.from_iterable(nested_labels))

    ## Loop over df_base
    results = []  # TODO: Preallocate?
    for row_i in range(df_base.shape[0]):
        ## Evaluate
        df_left = eval_df(
            model,
            tran_outer(
                DataFrame(
                    columns=var, data=-stencil + df_base[var].iloc[[row_i]].values
                ),
                df_base[var_fix].iloc[[row_i]],
            ),
            append=False,
        )

        df_right = eval_df(
            model,
            tran_outer(
                DataFrame(
                    columns=var, data=+stencil + df_base[var].iloc[[row_i]].values
                ),
                df_base[var_fix].iloc[[row_i]],
            ),
            append=False,
        )

        ## Compute differences
        res = (stepscale * (df_right[outputs] - df_left[outputs]).values).flatten()
        df_grad = DataFrame(columns=grad_labels, data=[res])

        results.append(df_grad)

    # Consolidate results
    df_res = concat(results).reset_index(drop=True)

    if append:
        return concat((df_base, df_res), axis=1)
    else:
        return df_res


ev_grad_fd = add_pipe(eval_grad_fd)

## Conservative quantile evaluation
# --------------------------------------------------
@curry
def eval_conservative(model, quantiles=None, df_det=None, append=True, skip=False):
    r"""Evaluates a given model at conservative input quantiles

    Uses model specifications to determine the "conservative" direction for each input, and evaluates the model at the desired quantile. Provided primarily for comparing UQ against pseudo-deterministic design criteria (del Rosario et al.; 2021).

    Note that if there is no conservative direction for the given input, the given quantile will be ignored and the median will automatically be selected.

    Args:
        model (gr.Model): Model to evaluate
        quantiles (numeric): lower quantile value(s) for conservative
            evaluation; can be single value for all inputs, array
            of values for each random variable, or None for default 0.01.
            values in [0, 0.5]
        df_det (DataFrame or None): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels. If provided model has no
            deterministic variables (model.n_var_det == 0), then df_det may
            equal None.
        append (bool): Append results to conservative inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Conservative evaluation or unevaluated design

    References:
        del Rosario, Zachary, Richard W. Fenrich, and Gianluca Iaccarino. "When Are Allowables Conservative?." AIAA Journal 59.5 (2021): 1760-1772.

    Examples::

        import grama as gr
        from grama.models import make_plate_buckle
        md = make_plate_buckle()
        # Evaluate at conservative input values
        md >> gr.ev_conservative(df_det="nom")

    """
    ## Check invariants
    invariants_eval_model(model, skip)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom"],
        acc_none=(model.n_var_det==0))

    ## Default behavior
    if quantiles is None:
        print("eval_conservative() using quantile default 0.01;")
        print("provide `quantiles` keyword for non-default behavior.")
        quantiles = [0.01] * model.n_var_rand

    ## Handle scalar vs vector quantiles
    try:
        len(quantiles)
    except TypeError:
        quantiles = [quantiles] * model.n_var_rand

    ## Modify quantiles for conservative directions
    quantiles = [
        0.5 + (0.5 - quantiles[i]) * model.density.marginals[model.var_rand[i]].sign
        for i in range(model.n_var_rand)
    ]
    quantiles = atleast_2d(quantiles)

    ## Draw samples
    df_pr = DataFrame(data=quantiles, columns=model.var_rand)
    ## Convert samples to desired marginals
    df_rand = model.density.pr2sample(df_pr)
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)
    if skip:
        ## Evaluation estimate
        runtime_msg = model.runtime_message(df_samp)
        print(runtime_msg)

        return df_samp
    return eval_df(model, df=df_samp, append=append)


ev_conservative = add_pipe(eval_conservative)

## Random sampling
# --------------------------------------------------
@curry
def eval_sample(model, n=None, df_det=None, seed=None, append=True, skip=False, comm=True, ind_comm=None):
    r"""Draw a random sample

    Evaluates a model with a random sample of the random model inputs. Generates outer product with deterministic levels (common random numbers) OR generates a sample fully-independent of deterministic levels (non-common random numbers).

    For more expensive models, it can be helpful to tune n to achieve a reasonable runtime. An even more effective approach is to use skip evaluation along with tran_sp() to evaluate a small, representative sample. (See examples below.)

    Args:
        model (gr.Model): Model to evaluate
        n (numeric): number of observations to draw
        df_det (DataFrame or None): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels. If provided model has no
            deterministic variables (model.n_var_det == 0), then df_det may
            equal None.
        seed (int): random seed to use
        append (bool): Append results to input values?
        skip (bool): Skip evaluation of the functions?
        comm (bool): Use common random numbers (CRN) across deterministic levels? CRN will tend to aid in the comparison of statistics across deterministic levels and enhance the convergence of stochastic optimization.
        ind_comm (str or None): Name of realization index column; not added if None

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    Examples::

        import grama as gr
        from grama.models import make_test
        DF = gr.Intention()

        # Simple random sample evaluation
        md = make_test()
        df = md >> gr.ev_sample(n=1e2, df_det="nom")
        df.describe()

        ## Use autoplot to visualize results
        (
            md
            >> gr.ev_sample(n=1e2, df_det="nom")
            >> gr.pt_auto()
        )

        ## Cantilever beam examples
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()

        ## Use the realization index to facilitate plotting
        # Try running this without the `group` aesthetic in `geom_line()`;
        # without the group the plot will not have multiple lines.
        (
            md_beam
            >> gr.ev_sample(
                n=20,
                df_det=gr.df_make(w=3, t=gr.linspace(2, 4, 100)),
                ind_comm="idx",
            )

            >> gr.ggplot(gr.aes("t", "g_stress"))
            + gr.geom_line(gr.aes(color="w", group="idx"))
        )

        ## Use iocorr to generate input/output correlation tile plot
        (
            md_beam
            >> gr.ev_sample(n=1e3, df_det="nom", skip=True)
            # Generate input/output correlation summary
            >> gr.tf_iocorr()
            # Visualize
            >> gr.pt_auto()
        )

        ## Use support points to reduce model runtime
        (
            md_beam
            # Generate large input sample but don't evaluate outputs
            >> gr.ev_sample(n=1e5, df_det="nom", skip=True)
            # Reduce to a smaller---but representative---sample
            >> gr.tf_sp(n=50)
            # Evaluate the outputs
            >> gr.tf_md(md_beam)
        )

        ## Estimate probabilities
        (
            md_beam
            # Generate large
            >> gr.ev_sample(n=1e5, df_det="nom")
            # Estimate probabilities of failure
            >> gr.tf_summarize(
                pof_stress=gr.mean(DF.g_stress <= 0),
                pof_disp=gr.mean(DF.g_disp <= 0),
            )
        )


    """
    ## Check invariants
    invariants_eval_model(model, skip)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom"],
        acc_none=(model.n_var_det==0))
    if n is None:
        raise ValueError("Must provide a valid n value.")

    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    ## Ensure sample count is int
    if not isinstance(n, Integral):
        print("eval_sample() is rounding n...")
        n = int(n)

    ## Draw realizations
    # Common random numbers
    if comm:
        df_rand = model.density.sample(n=n, seed=seed)
        if not ind_comm is None:
            df_rand[ind_comm] = df_rand.index
        df_samp = model.var_outer(df_rand, df_det=df_det)
    # Non-common random numbers
    else:
        df_rand = model.density.sample(n=n * df_det.shape[0], seed=seed)
        if not ind_comm is None:
            df_rand[ind_comm] = df_rand.index
        df_samp = concat(
            (df_rand, concat([df_det[model.var_det]]*n, axis=0).reset_index(drop=True)),
            axis=1,
        ).reset_index(drop=True)


    if skip:
        ## Evaluation estimate
        runtime_msg = model.runtime_message(df_samp)
        print(runtime_msg)

        ## Attach metadata
        with catch_warnings():
            simplefilter("ignore")
            df_samp._plot_info = {
                "type": "sample_inputs",
                "var": model.var_rand,
            }

        return df_samp


    df_res = eval_df(model, df=df_samp, append=append)
    ## Attach metadata
    with catch_warnings():
        simplefilter("ignore")
        df_res._plot_info = {
            "type": "sample_outputs",
            "var": model.var,
            "out": model.out,
        }

    return df_res


ev_sample = add_pipe(eval_sample)



__all__ = ["make_pipe_friction"]

from grama import cp_bounds, cp_function, cp_marginals, Model, df_make
from numpy import array
from math import sqrt, log10, pow, log
from scipy.optimize import bisect

def re_fcn(rho,u,d,mu,eps):
    # {rho,u,d,mu,eps}
    return rho*u*d/mu

def f_lam(rho,u,d,mu,eps):
    # {rho,u,d,mu,eps}
    return 64. / re_fcn(rho,u,d,mu,eps)

def colebrook(rho,u,d,mu,eps,f):
    # {rho,u,d,mu,eps}
    fs = sqrt(f); Re = re_fcn(rho,u,d,mu,eps)
    return 1 + 2.*fs*log10(eps/3.6/d + 2.51/Re/fs)

def f_tur(rho,u,d,mu,eps):
    return bisect(lambda f: colebrook(rho,u,d,mu,eps,f), 1e-5, 10)

Re_c = 3e3
def fcn_pipe(rho,u,d,mu,eps):
    Re = re_fcn(rho,u,d,mu,eps)
    if Re < Re_c:
        return f_lam(rho,u,d,mu,eps)
    else:
        return f_tur(rho,u,d,mu,eps)

def make_pipe_friction():
    r"""Pipe Friction Factor

    Approximate the pipe friction factor via simple equations. In the laminar regime ($Re < 3,000$ assumed) the analytic Poiseuille equation is used. Beyond $Re = 3,000$ the empirical expressions of Colebrook and White are used.

    Deterministic variables:
        d: Pipe diameter (m)
        eps: Pipe roughness lengthscale (m)

        rho: Fluid density (kg / m^3)
        u: Fluid bulk velocity (m / s)
        mu: Fluid dynamic viscosity (Pa * s)

    Random Variables: None

    Outputs:
        f: friction factor (-)

    References:
        White, *Fluid Mechanics*, McGraw-Hill, New York, 2011
        Nikuradse, *Laws of Flow in Rough Pipes*, National Advisory Committee for Aeronautics, Washington, D.C., 1950
        del Rosario, Lee, and Iaccarino, "Lurking Variable Detection via Dimensional Analysis," *SIAM/ASA J. Uncertainty Quantification*, 2019

    """
    md = (
        Model("Pipe Friction Factor")
        >> cp_function(
            fun=fcn_pipe,
            var=["rho", "u", "d", "mu", "eps"],
            out=["f"]
        )
        >> cp_bounds(
            rho=(1.0, 1.4),
            u=(1e-4, 1e+1),
            d=(1.3, 1.7),
            mu=(1.0e-5, 1.5e-5),
            eps=(0.5e-1, 2.0e-1),
        )
    )

    return md



""" Local linear models for use in other functions
"""
import numpy as np
from scipy.optimize import root_scalar
from scipy.special import xlogy
from scipy.spatial.distance import cdist
import scipy.linalg

__all__ = ['perplexity_bandwidth', 'local_linear_grads', 'local_linear']

def _compute_p1(M, perplexity):
	r""" The constant appearing in VC13, eq. 9
	"""
	#res = root_scalar(lambda x: np.log(min(np.sqrt(2*M), perplexity)) - 2*(1 - x)*np.log(M/(2*(1 - x))), bracket = [3./4,1 - 1e-14],)
	#p1 = res.root
	# Instead we solve in terms of x = 2*(1-p1)
	# x * log(N/x) = x * log(N) - x*log(x)
	# and xlogy(x, x) = x * log(x)
	fun = lambda x: x * np.log(M) - xlogy(x, x) - np.log(np.min([np.sqrt(2*M), perplexity]))

	try:
		res = root_scalar(fun, bracket = [0, 0.5])
		x = res.root
	except ValueError:
		if np.isclose(fun(0.5), 0):
			x = 0.5
		elif np.isclose(fun(0), 0):
			x = 0
		else:
			raise ValueError

	p1 = 1. - x/2.
	return p1


def log_entropy(beta, d):
	p = np.exp(-beta*d)
	sum_p = np.sum(p)
	# Shannon entropy H = np.sum(-p*np.log2(p))
	# More stable formula
	return beta*np.sum(p*d/sum_p) + np.log(sum_p)


def perplexity_bandwidth(d, perplexity):
	r"""Compute the bandwidth such that the exponential kernel has the desired perplexity

	Parameters
	----------
	d: array-like (M,)
		List of squared Euclidean distances
	perplexity: float, positive
		Target entropy of

	Returns
	-------
	bandwidth: float
		Bandwidth of Gaussian kernel (includes 1/2 factor)
	"""
	M = len(d)
	# TODO: Is perplexity necessarily in this interval
	#perplexity = min(M, perplexity)

	p1 = _compute_p1(M, perplexity)
	# Compute upper and lower bounds of beta from [VC13, eq. (7) (8)]
	# These are constants appearing the bounds
	dM = np.max(d)
	d1 = np.min(d[d>0])
	delta2 = d - d1
	delta2 = np.min(delta2[delta2>0])
	deltaM = dM - d1

	# lower bound VC13 (7)
	beta1 = max(M*np.log(M/perplexity)/((M-1)*deltaM), np.sqrt(np.log(M/perplexity)/(dM**4 - d1**4)))
	# upper bound VC13 (8)
	beta2 = 1/delta2*np.log(p1/(1-p1)*(M - 1))

	log_perplexity = np.log(perplexity)

	# TODO: Ideally we would use (root) Newton to find the optimal bandwidth
	# however Scipy doesn't support bounding intervals for its newton implementation.

	# TODO: There are some fancy initialization strategies that can be used
	# based on caching previous values; we ignore this here

	#print("beta1", beta1, log_entropy(beta1, d), "\nbeta2", beta2, log_entropy(beta2,d))


	# Compute bandwidth beta
	try:
		res = root_scalar(lambda beta: log_entropy(beta, d) - log_perplexity,
			bracket = [beta1, beta2],
			method = 'brenth',
			rtol = 1e-10)
		beta = res.root
	except ValueError as e:
		f1 = log_entropy(beta1, d) - log_perplexity
		f2 = log_entropy(beta2, d) - log_perplexity
		print("beta", beta1, "f", f1)
		print("beta", beta2, "f", f2)
		raise e
	return beta


def local_linear(X, fX, perplexity = None, bandwidth = None, Xt = None):
	r""" Construct local linear models at specified points

	In several dimension reduction settings we want to estimate gradients using only samples.
	If we had freedom to place these samples anywhere we wanted, we would use a finite difference
	approach.  As this is often not the case, we need some way to estimate gradients given a
	fixed and arbitrary set of data.

	Local linear models provide one approach for estimating the gradient.
	This approach constructs a local linear model centered around each :math:`\mathbf{x}_t`
	with weights depending on the distance between points

	.. math::
		\min_{a_0\in \mathbb{R}, \mathbf{a}\in \mathbb{R}^m}
			\sum_{i=1}^M [(a_0 + \mathbf{a}^\top \mathbf{x}_i) - f(\mathbf{x}_i)]^2 e^{-\beta_t \| \mathbf{x}_i - \mathbf{x}_t\|_2^2}.

	The choice of :math:`\beta_t` is critical. Here we provide two main options.
	By default, we choose :math:`\beta_t` for each :math:`\mathbf{x}_t`
	such that the perplexity corresponds to :math:`m+1`; other values of perplexity are avalible setting :code:`perplexity`.
	The other option is to specify the bandwidth :math:`\beta` explicitly.


	*Note* The cost of this method scales quadratically in the dimension of input space.

	Parameters
	----------
	X: array-like (M, m)
		Places where the function is evaluated
	fX: array-like (M,)
		Value of the function at those locations
	perplexity: None or float
		If None, defaults to m+1.
	bandwidth: None, 'xia' or positive float
		If specified, set the global bandwidth to the specified float.
		If 'xia', use the bandwidth selection heuristic of Xia mentioned
		in [Li18]_.

	Returns
	-------
	A: np.array (M, m+1)
		Matrix of coefficients of linear model;
		A[:,0] is the constant term and A[:,1:m+1] is the linear coefficients.
	"""

	M, m = X.shape
	fX = fX.flatten()
	assert len(fX) == M, "Number of function evaluations does not match number of samples"

	if Xt is None:
		Xt = X

	if perplexity is None and bandwidth is None:
		perplexity = min(m+1, M)
	if perplexity is not None:
		bandwidth = None
		assert perplexity >= 2 and perplexity < M, "Perplexity must be in the interval [2,M)"
	elif bandwidth is not None:
		perplexity = None
		if bandwidth == 'xia':
			# Bandwidth from Xia 2007, [Li18, eq. 11.5]
			bandwidth = 2.34*M**(-1./(max(M, 3) +6))


	# Storage for the gradients
	grads = np.zeros((len(Xt), m))
	Y = np.hstack([np.ones((M, 1)), X])

	# Coefficients in linear models
	A = np.zeros((M, m+1))
	for i, xi in enumerate(Xt):
		d = cdist(X, xi.reshape(1,-1), 'sqeuclidean').flatten()
		if perplexity:
			beta = perplexity_bandwidth(d, perplexity)
		else:
			beta = 0.5*bandwidth

		try:
			# Weights associated with each point
			sqrt_weights = np.exp(-0.5*beta*d).reshape(-1,1)
			#a, _, _, _ = np.linalg.lstsq(sqrt_weights*Y, sqrt_weights*fX.reshape(-1,1), rcond = None)
			a, _, _, _ = scipy.linalg.lstsq(sqrt_weights*Y, sqrt_weights*fX.reshape(-1,1), overwrite_a = True, overwrite_b = True)
			a = a.flatten()
		except np.linalg.LinAlgError:
			a = np.zeros(m+1)
		A[i,:] = a
	return A

def local_linear_grads(X, fX, perplexity = None, bandwidth = None, Xt = None):
	r""" Estimates the gradient from a local linear model
	"""
	A = local_linear(X, fX, perplexity = perplexity, bandwidth = bandwidth, Xt = Xt)
	return A[:,1:]



__all__ = ["make_plate_buckle"]

from grama import cp_bounds, cp_copula_gaussian, cp_vec_function, cp_marginals, \
    marg_fit, Model, df_make
from grama.data import df_stang
from numpy import pi


LOAD = 0.00128  # Applied load (kips)

## Reference geometry
THICKNESS = 0.06  # Plate thickness (in)
HEIGHT = 12.0  # Plate height    (in)


def make_plate_buckle():
    r"""Initialize a buckling plate model

    Variables (deterministic):
        w (in): Plate width
        h (in): Plate height
        t (in): Plate thickness
        m (-): Wavenumber
        L (kips): Applied (compressive) load;
            uniformly applied along top and bottom edges

    Variables (random):
        E (kips/in^2): Elasticity
        mu (-): Poisson's ratio

    Outputs:
        k_cr (-): Prefactor for buckling stress
        g_buckle (kips/in^2): Buckling limit state:
            critical stress - applied stress
    """
    md = (
        Model("Plate Buckling")
        >> cp_vec_function(
            fun=lambda df: df_make(
                k_cr=(df.m*df.h/df.w + df.w/df.m/df.h)**2
            ),
            var=["w", "h", "m"],
            out=["k_cr"],
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                g_buckle=df.k_cr * pi**2/12 * df.E / (1 - df.mu**2) * (df.t/df.h)**2
                - df.L / df.t / df.w
            ),
            var=["k_cr", "t", "h", "w", "E", "mu", "L"],
            out=["g_buckle"],
            name="limit state",
        )
        >> cp_bounds(
            t=(0.5 * THICKNESS, 2 * THICKNESS),
            h=(6, 18),
            w=(6, 18),
            m=(1, 5),
            L=(LOAD / 2, LOAD * 2),
        )
        >> cp_marginals(
            E=marg_fit("norm", df_stang.E),
            mu=marg_fit("beta", df_stang.mu),
        )
        >> cp_copula_gaussian(df_data=df_stang)
    )

    return md



import numpy as np

from scipy.stats import ortho_group
from scipy.linalg import orth
from scipy.spatial.distance import pdist

from functools import lru_cache


from .domain import Domain, TOL
from ..exceptions import SolverError, EmptyDomainException, UnboundedDomainException
from ..misc import merge
from ..quadrature import gauss


class EuclideanDomain(Domain):
	r""" Abstract base class for a Euclidean input domain

	This specifies a domain :math:`\mathcal{D}\subset \mathbb{R}^m`.

	"""

	def __init__(self, names = None):
		self._init_names(names)


	################################################################################
	# Naming of variables
	################################################################################

	@property
	def names(self):
		r""" Names associated with each of the variables (coordinates) of the domain
		"""
		try:
			return self._names
		except AttributeError:
			self._names = ['x%d' % i for i in range(len(self))]
			return self._names

	def _init_names(self, names):
		if names is None:
			return

		if isinstance(names, str):
			if len(self) == 1:
				self._names = [names]
			else:
				self._names = [names + ' %d' % (j+1,) for j in range(len(self)) ]
		else:
			assert len(self) == len(names), "Number of names must match dimension"
			self._names = names

	################################################################################
	# Properties of the domain
	################################################################################

	def __len__(self):
		r""" The dimension of the Euclidean space in which this domain lives.

		Returns
		-------
		m: int
			If the domain :math:`\mathcal{D} \subset \mathbb{R}^m`, this returns
			:math:`m`.
		"""
		return self._dimension




	def normalize(self, X):
		""" Given a points in the application space, convert it to normalized units

		Parameters
		----------
		X: np.ndarray((M,m))
			points in the domain to normalize
		"""
		try:
			X.shape
		except AttributeError:
			X = np.array(X)
		if len(X.shape) == 1:
			X = X.reshape(-1, len(self))
			return self._normalize(X).flatten()
		else:
			return self._normalize(X)

	def unnormalize(self, X_norm):
		""" Convert points from normalized units into application units

		Parameters
		----------
		X_norm: np.ndarray((M,m))
			points in the normalized domain to convert to the application domain

		"""
		if len(X_norm.shape) == 1:
			X_norm = X_norm.reshape(-1, len(self))
			return self._unnormalize(X_norm).flatten()
		else:
			return self._unnormalize(X_norm)


	def normalized_domain(self, **kwargs):
		""" Return a domain with units normalized corresponding to this domain
		"""
		return self._normalized_domain(**kwargs)


	################################################################################
	# Simple properties
	################################################################################

	@property
	def lb(self): return -np.inf*np.ones(len(self))

	@property
	def ub(self): return np.inf*np.ones(len(self))

	@property
	def A(self): return np.zeros((0, len(self)))

	@property
	def b(self): return np.zeros((0,))

	@property
	def A_aug(self):
		r""" Linear inequalities augmented with bound constraints as well
		"""
		I = np.eye(len(self))
		Ilb = np.isfinite(self.lb)
		Iub = np.isfinite(self.ub)
		return np.vstack([self.A, -I[Ilb,:], I[Iub,:]])

	@property
	def b_aug(self):
		Ilb = np.isfinite(self.lb)
		Iub = np.isfinite(self.ub)
		return np.hstack([self.b, -self.lb[Ilb], self.ub[Iub]])

	@property
	def A_eq(self): return np.zeros((0, len(self)))

	@property
	def b_eq(self): return np.zeros((0,))

	@property
	def Ls(self): return []

	@property
	def ys(self): return []

	@property
	def rhos(self): return []

	@property
	def lb_norm(self):
		lb_norm = self.normalize(self.lb)
		I = ~np.isfinite(self.lb)
		lb_norm[I] = -np.inf
		return lb_norm

	@property
	def ub_norm(self):
		ub_norm = self.normalize(self.ub)
		I = ~np.isfinite(self.ub)
		ub_norm[I] = np.inf
		return ub_norm

	@property
	def A_norm(self):
		D = self._unnormalize_der()
		return self.A.dot(D)

	@property
	def b_norm(self):
		c = self._center()
		return self.b - self.A.dot(c)

	@property
	def A_eq_norm(self):
		D = self._unnormalize_der()
		return self.A_eq.dot(D)

	@property
	def b_eq_norm(self):
		c = self._center()
		return self.b_eq - self.A_eq.dot(c)

	@property
	def Ls_norm(self):
		D = self._unnormalize_der()
		return [ L.dot(D) for L in self.Ls]

	@property
	def ys_norm(self):
		c = self._center()
		return [y - c for y in self.ys]

	@property
	def rhos_norm(self):
		return self.rhos

	################################################################################
	# Meta properties
	################################################################################

	def _is_linquad_domain(self):
		return True

	def _is_linineq_domain(self):
		return len(self.Ls) == 0

	def _is_box_domain(self):
		return len(self.Ls) == 0 and len(self.b) == 0 and len(self.b_eq) == 0


	# These are the lower and upper bounds to use for normalization purposes;
	# they do not add constraints onto the domain.
	@property
	def norm_lb(self):
		r"""Lower bound used for normalization purposes; does not constrain the domain
		"""
		try:
			return self._norm_lb
		except AttributeError:
			self._norm_lb = -np.inf*np.ones(len(self))
			for i in range(len(self)):
				ei = np.zeros(len(self))
				ei[i] = 1
				if np.isfinite(self.lb[i]):
					self._norm_lb[i] = self.lb[i]
					# This ensures normalization maintains positive orientation
					if np.isfinite(self.ub[i]):
						self._norm_lb[i] = min(self._norm_lb[i], self.ub[i])
				else:
					try:
						x_corner = self.corner(-ei)
						self._norm_lb[i] = x_corner[i]
					except (SolverError, UnboundedDomainException):
						self._norm_lb[i] = -np.inf

			return self._norm_lb

	@property
	def norm_ub(self):
		r"""Upper bound used for normalization purposes; does not constrain the domain
		"""
		try:
			return self._norm_ub
		except AttributeError:
			# Temporarly disable normalization

			# Note: since this will be called by corner, we need to
			# choose a reasonable value to initialize this property, which
			# will be used until the remainder of the corner calls are made
			self._norm_ub = np.inf*np.ones(len(self))
			for i in range(len(self)):
				ei = np.zeros(len(self))
				ei[i] = 1
				if np.isfinite(self.ub[i]):
					self._norm_ub[i] = self.ub[i]
					# This ensures normalization maintains positive orientation
					if np.isfinite(self.ub[i]):
						self._norm_ub[i] = max(self._norm_ub[i], self.lb[i])
				else:
					try:
						x_corner = self.corner(ei)
						self._norm_ub[i] = x_corner[i]
					except (SolverError, UnboundedDomainException):
						self._norm_ub[i] = np.inf

			return self._norm_ub

	################################################################################
	# Normalization functions
	################################################################################

	def isnormalized(self):
		return np.all( (~np.isfinite(self.norm_lb)) | (self.norm_lb == -1.) ) and np.all( (~np.isfinite(self.norm_ub)) | (self.norm_ub == 1.) )

	def _normalize_der(self):
		"""Derivative of normalization function"""

		slope = np.ones(len(self))
		I = (self.norm_ub != self.norm_lb) & np.isfinite(self.norm_lb) & np.isfinite(self.norm_ub)
		slope[I] = 2.0/(self.norm_ub[I] - self.norm_lb[I])
		return np.diag(slope)

	def _unnormalize_der(self):
		slope = np.ones(len(self))
		I = (self.norm_ub != self.norm_lb) & np.isfinite(self.norm_lb) & np.isfinite(self.norm_ub)
		slope[I] = (self.norm_ub[I] - self.norm_lb[I])/2.0
		return np.diag(slope)

	def _center(self):
		c = np.zeros(len(self))
		I = np.isfinite(self.norm_lb) & np.isfinite(self.norm_ub)
		c[I] = (self.norm_lb[I] + self.norm_ub[I])/2.0
		return c

	def _normalize(self, X):
		c = self._center()
		D = self._normalize_der()
		X_norm = D.dot( (X - c.reshape(1,-1)).T ).T
		return X_norm

	def _unnormalize(self, X_norm, **kwargs):
		c = self._center()
		Dinv = self._unnormalize_der()
		X = Dinv.dot(X_norm.T).T + c.reshape(1,-1)
		return X

	################################################################################
	# Bound checking
	################################################################################
	def _isinside_bounds(self, X, tol = None):
		X = np.array(X)
		if tol is None: tol = self.tol
		#lb_check = np.array([np.all(x >= self.lb-tol) for x in X], dtype = np.bool)
		#ub_check = np.array([np.all(x <= self.ub+tol) for x in X], dtype = np.bool)
		lb_check = np.ones(len(X), dtype = np.bool)
		ub_check = np.ones(len(X), dtype = np.bool)
		for i in range(len(self)):
			if np.isfinite(self.lb[i]):
				lb_check &= X[:,i] >= self.lb[i] - tol
			if np.isfinite(self.ub[i]):
				ub_check &= X[:,i] <= self.ub[i] + tol
		#print("bounds check", lb_check & ub_check, self.names)
		return lb_check & ub_check

	def _isinside_ineq(self, X, tol = None):
		if tol is None: tol = self.tol
		return np.array([np.all(np.dot(self.A, x) <= self.b + tol) for x in X], dtype = np.bool)

	def _isinside_eq(self, X, tol = None):
		if tol is None: tol = self.tol
		return np.array([np.all( np.abs(np.dot(self.A_eq, x) - self.b_eq) < tol) for x in X], dtype = np.bool)

	def _isinside_quad(self, X, tol = None):
		"""check that points are inside quadratic constraints"""
		if tol is None: tol = self.tol
		inside = np.ones(X.shape[0],dtype = np.bool)
		for L, y, rho in zip(self.Ls, self.ys, self.rhos):
			diff = X - np.tile(y.reshape(1,-1), (X.shape[0],1))
			Ldiff = L.dot(diff.T).T
			Ldiff_norm = np.sum(Ldiff**2,axis=1)
			inside = inside & (np.sqrt(Ldiff_norm) <= rho + tol)
		return inside


	################################################################################
	# Extent functions
	################################################################################

	def _extent_bounds(self, x, p):
		"""Check the extent from the box constraints"""
		alpha = np.inf

		# If on the boundary, the direction needs to point inside the domain
		# otherwise we cannot move
		if np.any(p[self.lb == x] < 0):
			return 0.
		if np.any(p[self.ub == x] > 0):
			return 0.

		# To prevent divide-by-zero we ignore directions we are not moving in
		I = np.nonzero(p)

		# Check upper bounds
		y = (self.ub - x)[I]/p[I]
		if np.any(y>0):
			alpha = min(alpha, np.min(y[y>0]))

		# Check lower bounds
		y = (self.lb - x)[I]/p[I]
		if np.any(y>0):
			alpha = min(alpha, np.min(y[y>0]))

		return alpha

	def _extent_ineq(self, x, p):
		""" check the extent from the inequality constraints """
		alpha = np.inf
		# positive extent
		with np.errstate(divide = 'ignore'):
			y = (self.b - np.dot(self.A, x)	)/np.dot(self.A, p)
		if np.sum(y>0) > 0:
			alpha = min(alpha, np.min(y[y>0]))

		return alpha

	def _extent_quad(self, x, p):
		""" check the extent from the quadratic constraints"""
		alpha = np.inf
		for L, y, rho in zip(self.Ls, self.ys, self.rhos):
			Lp = L.dot(p)
			Lxy = L.dot(x - y)
			# Terms in quadratic formula a alpha^2 + b alpha + c
			a = Lp.T.dot(Lp)
			b = 2*Lp.T.dot(Lxy)
			c = Lxy.T.dot(Lxy) - rho**2

			roots = np.roots([a,b,c])
			real_roots = roots[np.isreal(roots)]
			pos_roots = real_roots[real_roots>=0]
			if len(pos_roots) > 0:
				alpha = min(alpha, min(pos_roots))
		return alpha



import numpy as np
import pandas as pd
import unittest

from context import grama as gr
from context import data

##==============================================================================
## join test functions
##==============================================================================


class TestJoins(unittest.TestCase):
    def setUp(self):
        self.dfA = pd.DataFrame({"x1": ["A", "B", "C"], "x2": [1, 2, 3]})
        self.dfB = pd.DataFrame({"x1": ["A", "B", "D"], "x3": [True, False, True]})
        self.dfC = pd.DataFrame({"x1": ["B", "C", "D"], "x2": [2, 3, 4]})

    def test_inner_join(self):
        ab = pd.DataFrame({"x1": ["A", "B"], "x2": [1, 2], "x3": [True, False]})

        c = self.dfA >> gr.tf_inner_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))

    def test_outer_join(self):
        ab = pd.DataFrame(
            {
                "x1": ["A", "B", "C", "D"],
                "x2": [1, 2, 3, np.nan],
                "x3": [True, False, np.nan, True],
            }
        )

        c = self.dfA >> gr.tf_outer_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))
        c = self.dfA >> gr.tf_full_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))

    def test_left_join(self):
        ab = pd.DataFrame(
            {"x1": ["A", "B", "C"], "x2": [1, 2, 3], "x3": [True, False, np.nan]}
        )

        c = self.dfA >> gr.tf_left_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))

    def test_right_join(self):
        ab = pd.DataFrame(
            {"x1": ["A", "B", "D"], "x2": [1, 2, np.nan], "x3": [True, False, True]}
        )

        c = self.dfA >> gr.tf_right_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))

    def test_semi_join(self):
        ab = pd.DataFrame({"x1": ["A", "B"], "x2": [1, 2]})

        c = self.dfA >> gr.tf_semi_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))

    def test_anti_join(self):
        ab = pd.DataFrame({"x1": ["C"], "x2": [3]}, index=[2])

        c = self.dfA >> gr.tf_anti_join(self.dfB, by="x1")
        self.assertTrue(c.equals(ab))

    ##==============================================================================
    ## set operation (row join) test functions
    ##==============================================================================

    def test_union(self):
        ac = pd.DataFrame(
            {"x1": ["A", "B", "C", "D"], "x2": [1, 2, 3, 4]}, index=[0, 1, 2, 2]
        )

        d = self.dfA >> gr.tf_union(self.dfC)
        self.assertTrue(d.equals(ac))

    def test_intersect(self):
        ac = pd.DataFrame({"x1": ["B", "C"], "x2": [2, 3]})

        d = self.dfA >> gr.tf_intersect(self.dfC)
        self.assertTrue(d.equals(ac))

    def test_set_diff(self):
        ac = pd.DataFrame({"x1": ["A"], "x2": [1]})

        d = self.dfA >> gr.tf_set_diff(self.dfC)
        self.assertTrue(d.equals(ac))

    ##==============================================================================
    ## bind rows, cols
    ##==============================================================================

    def test_bind_rows(self):
        inner = pd.DataFrame({"x1": ["A", "B", "C", "A", "B", "D"]})
        outer = pd.DataFrame(
            {
                "x1": ["A", "B", "C", "A", "B", "D"],
                "x2": [1, 2, 3, np.nan, np.nan, np.nan],
                "x3": [np.nan, np.nan, np.nan, True, False, True],
            }
        )
        ab_inner = self.dfA >> gr.tf_bind_rows(self.dfB, join="inner")
        ab_outer = self.dfA >> gr.tf_bind_rows(self.dfB, join="outer")
        self.assertTrue(inner.equals(ab_inner.reset_index(drop=True)))
        self.assertTrue(outer.equals(ab_outer.reset_index(drop=True)))

    def test_bind_cols(self):
        dfB = self.dfB.copy()
        dfB.columns = ["x3", "x4"]
        df = pd.DataFrame(
            {
                "x1": ["A", "B", "C"],
                "x2": [1, 2, 3],
                "x3": ["A", "B", "D"],
                "x4": [True, False, True],
            }
        )
        d = self.dfA >> gr.tf_bind_cols(dfB)
        self.assertTrue(df.equals(d))



import numpy as np
import pandas as pd
import unittest

from context import grama as gr
from context import models


class TestPlumbing(unittest.TestCase):
    """Test implementation of pipe-shortcut functions

    Note that these do not check correctness of the original function,
    just that the piped version matches the non-piped version.
    """

    def setUp(self):
        self.model_default = models.make_test()
        self.df_test = models.df_test_input

    ## Evaluations
    # --------------------------------------------------

    def test_ev_df(self):
        """Check ev_df()
        """
        df_res = gr.eval_df(self.model_default, df=self.df_test)

        self.assertTrue(
            gr.df_equal(df_res, self.model_default >> gr.ev_df(df=self.df_test))
        )

    def test_ev_nominal(self):
        """Check ev_nominal()
        """
        df_res = gr.eval_nominal(self.model_default, df_det="nom")

        self.assertTrue(
            gr.df_equal(df_res, self.model_default >> gr.ev_nominal(df_det="nom"))
        )

    def test_ev_grad_fd(self):
        """Check ev_grad_fd()
        """
        df_res = gr.eval_grad_fd(self.model_default, df_base=self.df_test)

        self.assertTrue(
            gr.df_equal(
                df_res, self.model_default >> gr.ev_grad_fd(df_base=self.df_test)
            )
        )

    def test_ev_conservative(self):
        """Check ev_conservative()
        """
        df_res = gr.eval_conservative(self.model_default, df_det="nom")

        self.assertTrue(
            gr.df_equal(df_res, self.model_default >> gr.ev_conservative(df_det="nom"))
        )

    def test_ev_sample(self):
        """Check ev_sample()
        """
        df_res = gr.eval_sample(self.model_default, n=1, seed=101, df_det="nom")

        self.assertTrue(
            gr.df_equal(
                df_res,
                self.model_default >> gr.ev_sample(seed=101, n=1, df_det="nom")
            )
        )

    def test_ev_sinews(self):
        """Check ev_sinews()
        """
        df_res = gr.eval_sinews(self.model_default, seed=101, df_det="nom")

        self.assertTrue(
            gr.df_equal(
                df_res, self.model_default >> gr.ev_sinews(seed=101, df_det="nom")
            )
        )

    ## Fittings
    # --------------------------------------------------

    ## TODO

    ## Compositions
    # --------------------------------------------------

    ## TODO

    ## Transforms
    # --------------------------------------------------

    ## TODO


## Run tests
if __name__ == "__main__":
    unittest.main()



__all__ = [
    "abs",
    "sin",
    "cos",
    "tan",
    "log",
    "exp",
    "sqrt",
    "pow",
    "floor",
    "ceil",
    "round",
    "as_int",
    "as_float",
    "as_str",
    "as_factor",
    "as_numeric",
    "fct_reorder",
    "fillna",
    "qnorm",
    "pnorm",
    "dnorm",
    "pareto_min",
    "stratum_min",
    "qqvals",
    "linspace",
    "logspace",
    "consec",
]

from grama import make_symbolic, marg_fit
from numpy import array, median, zeros, ones, NaN, arange
from numpy import argsort, argmin, argmax
from numpy import any as npany
from numpy import all as npall
from numpy import abs as npabs
from numpy import sin as npsin
from numpy import cos as npcos
from numpy import tan as nptan
from numpy import log as nplog
from numpy import exp as npexp
from numpy import sqrt as npsqrt
from numpy import power as nppower
from numpy import floor as npfloor
from numpy import ceil as npceil
from numpy import round as npround
from numpy import linspace as nplinspace
from numpy import logspace as nplogspace
from pandas import Categorical, Series, to_numeric, concat
from scipy.stats import norm, rankdata


# --------------------------------------------------
# Mutation helpers
# --------------------------------------------------
# Numeric
# -------------------------
@make_symbolic
def floor(x):
    r"""Round downwards to nearest integer
    """
    return npfloor(x)


@make_symbolic
def ceil(x):
    r"""Round upwards to nearest integer
    """
    return npceil(x)


@make_symbolic
def round(x, decimals=0):
    r"""Round value to desired precision

    Args:
        x (float or iterable of floats): Value(s) to round
        decimals (int): Number of decimal places for rounding; decimals=0 rounds to integers

    """
    return npround(x, decimals=decimals)


@make_symbolic
def abs(x):
    r"""Absolute value
    """
    return npabs(x)


@make_symbolic
def sin(x):
    r"""Sine
    """
    return npsin(x)


@make_symbolic
def cos(x):
    r"""Cosine
    """
    return npcos(x)

@make_symbolic
def tan(x):
    r"""Tangent
    """
    return nptan(x)

@make_symbolic
def log(x):
    r"""(Natural) log
    """
    return nplog(x)


@make_symbolic
def exp(x):
    r"""Exponential (e-base)
    """
    return npexp(x)


@make_symbolic
def sqrt(x):
    r"""Square-root
    """
    return npsqrt(x)


@make_symbolic
def pow(x, p):
    r"""Power

    Usage:
        q = pow(x, p) := x ^ p

    Arguments:
        x = base
        p = exponent
    """
    return nppower(x, p)


# Casting
# -------------------------
@make_symbolic
def as_int(x):
    r"""Cast to integer
    """
    return x.astype(int)


@make_symbolic
def as_float(x):
    r"""Cast to float
    """
    return x.astype(float)


@make_symbolic
def as_str(x):
    r"""Cast to string
    """
    return x.astype(str)


@make_symbolic
def as_factor(x, categories=None, ordered=True, dtype=None):
    r"""Cast to factor

    Args:
        x (pd.Series): Column to convert
        categories (list): Categories (levels) of factor (Optional)
        ordered (boolean): Order the factor?
    """
    return Categorical(x, categories=categories, ordered=ordered, dtype=dtype)


@make_symbolic
def as_numeric(x):
    r"""Cast to numeric
    """
    return to_numeric(x, errors="coerce")


# Distributions
# -------------------------
@make_symbolic
def qnorm(x):
    r"""Normal quantile function (inverse CDF)
    """
    return norm.ppf(x)


@make_symbolic
def dnorm(x):
    r"""Normal probability density function (PDF)
    """
    return norm.pdf(x)


@make_symbolic
def pnorm(x):
    r"""Normal cumulative distribution function (CDF)
    """
    return norm.cdf(x)


# Pareto frontier calculation
# -------------------------
@make_symbolic
def pareto_min(*args):
    r"""Determine if observation is a Pareto point

    Find the Pareto-efficient points that minimize the provided features.

    Args:
        xi (iterable OR gr.Intention()): Feature to minimize; use -X to maximize

    Returns:
        np.array of boolean: Indicates if observation is Pareto-efficient

    """
    # Check invariants
    lengths = map(len, args)
    if len(set(lengths)) > 1:
        raise ValueError("All arguments to pareto_min must be of equal length")

    # Compute pareto points
    costs = array([*args]).T
    is_efficient = ones(costs.shape[0], dtype=bool)
    for i, c in enumerate(costs):
        is_efficient[i] = npall(npany(costs[:i] > c, axis=1)) and npall(
            npany(costs[i + 1 :] > c, axis=1)
        )

    return is_efficient

# Shell number calculation
# -------------------------
@make_symbolic
def stratum_min(*args, max_depth=10):
    r"""Compute Pareto stratum number

    Compute the Pareto stratum number for a given dataset.

    Args:
        xi (iterable OR gr.Intention()): Feature to minimize; use -X to maximize
        max_depth (int): Maximum depth for recursive computation; stratum numbers exceeding this value will not be computed and will be flagged as NaN.

    Returns:
        np.array of floats: Pareto stratum number

    References:
        del Rosario, Rupp, Kim, Antono, and Ling "Assessing the frontier: Active learning, model accuracy, and multi-objective candidate discovery and optimization" (2020) J. Chem. Phys.
    """
    # Check invariants
    lengths = map(len, args)
    if len(set(lengths)) > 1:
        raise ValueError("All arguments to stratum_min must be of equal length")

    # Set default as NaN
    costs = array([*args]).T
    n = costs.shape[0]
    stratum = ones(n)
    stratum[:] = NaN

    # Successive computation of stratum numbers
    active = ones(n, dtype=bool)
    idx_all = arange(n, dtype=int)

    i = 1
    while any(active) and (i <= max_depth):
        idx = idx_all[active]
        pareto = pareto_min(costs[idx].T)
        stratum[idx[pareto]] = i
        active[idx[pareto]] = False

        i += 1

    return stratum


# Factors
# -------------------------
@make_symbolic
def fct_reorder(f, x, fun=median):
    r"""Reorder a factor on another variable

    Args:
        f (iterable OR DataFrame column): factor to reorder
        x (iterable OR DataFrame column): variable on which to reorder; specify aggregation method with fun
        fun (function): aggregation function for reordering, default=median

    Returns:
        Categorical: Iterable with levels sorted according to x

    Examples::

        import grama as gr
        from grama.data import df_diamonds
        DF = gr.Intention()
        (
            df_diamonds
            >> gr.tf_mutate(cut=gr.fct_reorder(DF.cut, DF.price))
            >> gr.tf_group_by(DF.cut)
            >> gr.tf_summarize(max=gr.colmax(DF.price), mean=gr.mean(DF.price))
        )
    """

    # Get factor levels
    levels = array(list(set(f)))
    # Compute given fun over associated values
    values = zeros(len(levels))
    for i in range(len(levels)):
        mask = f == levels[i]
        values[i] = fun(x[mask])
    # Sort according to computed values
    return as_factor(f, categories=levels[argsort(values)], ordered=True)


# Pandas helpers
# -------------------------
@make_symbolic
def fillna(*args, **kwargs):
    r"""Wrapper for pandas Series.fillna

    (See below for Pandas documentation)

    Examples::

        import grama as gr
        X = gr.Intention()
        df = gr.df_make(x=[1, gr.NaN], y=[2, 3])
        df_filled = (
            df
            >> gr.tf_mutate(x=gr.fillna(X.x, 0))
        )
    """
    return Series.fillna(*args, **kwargs)


fillna.__doc__ = fillna.__doc__ + Series.fillna.__doc__

# Q-Q Plot Helper
# -------------------------
@make_symbolic
def qqvals(x, dist=None, marg=None):
    r"""Generate theoretical quantiles

    Generate theoretical quantiles for a Q-Q plot. Can provide either a pre-defined Marginal object or the name of a distribution to fit.

    Arguments:
        x (array-like or gr.Intention()): Target observations

    Keyword Arguments:
        dist (str or None): Name of scipy distribution to fit; see gr.valid_dist for list of valid distributions
        marg (gr.Marginal() or None): Pre-fitted marginal

    Returns:
        Series: Theoretical quantiles, matched in order with target observations

    References:
        Filliben, J. J., "The Probability Plot Correlation Coefficient Test for Normality" (1975) Technometrics. DOI: 10.1080/00401706.1975.10489279

    Examples::

        import grama as gr
        from grama.data import df_shewhart
        DF = gr.Intention()

        ## Make a Q-Q plot
        (
            df_shewhart
            >> gr.tf_mutate(q=gr.qqvals(DF.tensile_strength, dist="norm"))
            >> gr.ggplot(gr.aes("q", "tensile_strength"))
            + gr.geom_abline(intercept=0, slope=1, linetype="dashed")
            + gr.geom_point()
        )

        ## Fit a marginal, use in Q-Q plot
        mg_tys = gr.marg_fit("lognorm", df_shewhart.tensile_strength, floc=0)
        # Plot to assess the fit
        (
            df_shewhart
            >> gr.tf_mutate(q=gr.qqvals(DF.tensile_strength, marg=mg_tys))
            >> gr.ggplot(gr.aes("q", "tensile_strength"))
            + gr.geom_abline(intercept=0, slope=1, linetype="dashed")
            + gr.geom_point()
        )

    """
    # Check invariants
    if (marg is None) and (dist is None):
        raise ValueError(
            "Must provide one of marg or dist (exclusively)."
        )
    if (marg is not None) and (dist is not None):
        raise ValueError(
            "Must provide either marg or dist (exclusively)."
        )

    # Handle marginal input
    if (dist is not None):
        marg = marg_fit(dist, x)

    # Get sorted probability values
    n = len(x)
    i = rankdata(x, method="ordinal")
    # Filliben order statistic medians
    p = (i - 0.3175) / (n + 0.365)
    p[argmax(x)] = 0.5**(1/n)
    p[argmin(x)] = 1 - 0.5**(1/n)

    return marg.q(p)


# Array constructors
# -------------------------
@make_symbolic
def linspace(a, b, n=100, **kwargs):
    r"""Linearly-spaced values

    Create an array of linearly-spaced values. Accepts keyword arguments for numpy.linspace.

    Arguments:
        a (numeric): Smallest value
        b (numeric): Largest value
        n (int): Number of points

    Returns:
        numpy array: Array of requested values

    Notes:
        This is a symbolic alias for np.linspace(); you can use this in pipe-enabled functions.

    Examples::

        import grama as gr
        from grama.data import df_stang
        DF = gr.Intention()
        (
            df_stang
            >> gr.tf_mutate(c=gr.linspace(0, 1, gr.n(DF.index)))
        )

    """
    return nplinspace(a, b, num=n, **kwargs)

@make_symbolic
def logspace(a, b, n=100, **kwargs):
    r"""Logarithmically-spaced values

    Create an array of logarithmically-spaced values. Accepts keyword arguments for numpy.logspace.

    Arguments:
        a (numeric): Smallest value
        b (numeric): Largest value
        n (int): Number of points

    Returns:
        numpy array: Array of requested values

    Notes:
        This is a symbolic alias for np.logspace(); you can use this in pipe-enabled functions.

    Examples::

        import grama as gr
        from grama.data import df_stang
        DF = gr.Intention()
        (
            df_stang
            >> gr.tf_mutate(c=gr.logspace(0, 1, gr.n(DF.index)))
        )

    """
    return nplogspace(a, b, num=n, **kwargs)


@make_symbolic
def consec(series, i=3):
    r"""Flag consecutive runs of True values

    Flag as True all entries associated with a "run" of True values. A "run" is
    defined as a user-defined count (given by i) of consecutive True values.

    Args:
        series (pd.Series): Series of boolean values
        i (int): User-defined count to define a run

    Returns:
        series: Boolean series flagging consecutive True values

    """
    # Pad the series
    s = concat((
        Series([False] * i),
        series,
        Series([False] * i),
    ))
    # Find the runs
    run = s.rolling(i, center=True).sum() \
           .rolling(i+1, center=True).max() \
           [i:-i] >= i
    # Touches run and is True
    return run & series



from __future__ import print_function
import numpy as np
import textwrap
import inspect

#from .domains import Domain

from .misc import merge


__all__ = ['BaseFunction']

class BaseFunction(object):
	r""" Abstract base class for functions

	"""
	def eval(self, X, **kwargs):
		return self.__call__(X, return_grad = False)

	def grad(self, X):
		return self.__call__(X, return_grad = True)[1]

	def hessian(self, X):
		raise NotImplementedError

	def __call__(self, X, return_grad = False, **kwargs):
		if return_grad:
			return self.eval(X, **kwargs), self.grad(X)
		else:
			return self.eval(X, **kwargs)

	def predict(self, X):
		r""" Alias of __call__ to match scikit learn API
		"""
		return self.__call__(X)



"""Ridge function approximation from function values"""
# (c) 2017 Jeffrey M. Hokanson (jeffrey@hokanson.us)

import numpy as np
import scipy.linalg
import scipy.special
import warnings
from copy import deepcopy, copy

from .domains import Domain, BoxDomain
from .function import BaseFunction
from .subspace import SubspaceBasedDimensionReduction
from .ridge import RidgeFunction
from .basis import *
from .gn import gauss_newton
# from .seqlp import sequential_lp
from .exceptions import UnderdeterminedException
from .initialization import initialize_subspace
from .poly import PolynomialFunction


class PolynomialRidgeFunction(RidgeFunction):
	r""" A polynomial ridge function
	"""
	def __init__(self, basis, coef, U):
		self.basis = basis
		self.coef = np.copy(coef)
		self._U = np.array(U)
		self.domain = None


	def V(self, X, U = None):
		if U is None: U = self.U
		X = np.array(X)
		Y = (U.T @ X.T).T
		return self.basis.V(Y)

	def DV(self, X, U = None):
		if U is None: U = self.U

		Y = (U.T @ X.T).T
		return self.basis.DV(Y)

	def DDV(self, X, U = None):
		if U is None: U = self.U
		Y = (U.T @ X.T).T
		return self.basis.DDV(Y)

	def eval(self, X):
		Vc = self.V(X) @ self.coef
		return Vc

	def grad(self, X):
		if len(X.shape) == 1:
			one_d = True
			X = X.reshape(1,-1)
		else:
			one_d = False

		DV = self.DV(X)
		# Compute gradient on projected space
		Df = np.tensordot(DV, self.coef, axes = (1,0))
		# Inflate back to whole space
		Df = Df.dot(self.U.T)
		if one_d:
			return Df.reshape(X.shape[1])
		else:
			return Df

	def hessian(self, X):
		if len(X.shape) == 1:
			one_d = True
			X = X.reshape(1,-1)
		else:
			one_d = False

		DDV = self.DDV(X)
		DDf = np.tensordot(DDV, self.coef, axes = (1,0))
		# Inflate back to proper dimensions
		DDf = np.tensordot(np.tensordot(DDf, self.U, axes = (2,1)) , self.U, axes = (1,1))
		if one_d:
			return DDf.reshape(X.shape[1], X.shape[1])
		else:
			return DDf

	@property
	def profile(self):
		return PolynomialFunction(self.basis, self.coef)



################################################################################
# Two types of custom errors raised by PolynomialRidgeApproximation
################################################################################
class UnderdeterminedException(Exception):
	pass

class IllposedException(Exception):
	pass


def orth(U):
	""" Orthgonalize, but keep directions"""
	U, R = np.linalg.qr(U, mode = 'reduced')
	U = np.dot(U, np.diag(np.sign(np.diag(R))))
	return U

def two_norm_fit(A,b):
	r""" solve 2-norm linear optimization problem

	.. math::

		\min_{x} \| \mathbf{A} \mathbf{x} - \mathbf{b}\|_2

	"""
	return scipy.linalg.lstsq(A, b)[0]

class PolynomialRidgeApproximation(PolynomialRidgeFunction):
	r""" Constructs a ridge approximation using a total degree approximation

	Given a basis of total degree polynomials :math:`\lbrace \psi_j \rbrace_{j=1}^N`
	on :math:`\mathbb{R}^n`, this class constructs a polynomial ridge function
	that minimizes the mismatch on a set of points :math:`\lbrace \mathbf{x}_i\rbrace_{i=1}^M \subset \mathbb{R}^m`
	in a :math:`p`-norm:

	.. math::

		\min_{\mathbf{U} \in \mathbb{R}^{m\times n}, \  \mathbf{U}^\top \mathbf{U} = \mathbf I, \
			\mathbf{c}\in \mathbb{R}^N }
			\sqrt[p]{ \sum_{i=1}^M  \left|f(\mathbf{x}_i) -
				\sum_{j=1}^N c_j \psi_j(\mathbf{U}^\top \mathbf{x}_i) \right|^p}

	This approach assumes :math:`\mathbf{U}` is an element of the Grassmann manifold
	obeying the orthogonality constraint.

	For the 2-norm (:math:`p=2`) this implementation uses Variable Projection following [HC18]_
	to remove the solution of the linear coefficients :math:`\mathbf{c}`,
	leaving an optimization problem posed over the Grassmann manifold alone.

	For both the 1-norm and the :math:`\infty`-norm,
	this implementation uses a sequential linear program with a trust region
	coupled with a nonlinear trajectory through the search space.

	Parameters
	----------
	degree: int
		Degree of polynomial

	subspace_dimension: int
		Dimension of the low-dimensional subspace associated with the ridge approximation.

	basis: ['legendre', 'monomial', 'chebyshev', 'laguerre', 'hermite']
		Basis for polynomial representation

	norm: [1, 2, np.inf, 'inf']
		Norm in which to evaluate the mismatch between the ridge approximation and the data

	scale: bool (default:True)
		Scale the coordinates along the ridge to ameliorate ill-conditioning

	bound: [None, 'lower', 'upper']
		If 'lower' or 'upper' construct a lower or upper bound

	rotate: bool
		If True, rotate the U matrix to align to the active subspace with average increasing gradients

	References
	----------
	.. [HC18] J. M. Hokanson and Paul G. Constantine.
		Data-driven Polynomial Ridge Approximation Using Variable Projection.
		SIAM J. Sci. Comput. Vol 40, No 3, pp A1566--A1589, DOI:10.1137/17M1117690.
	"""

	def __init__(self, degree, subspace_dimension, basis = 'legendre',
		norm = 2, n_init = 1, scale = True, keep_data = True, domain = None,
		bound = None, rotate = True, **kwargs):

		self.kwargs = kwargs
		self.rotate = rotate
		assert isinstance(degree, int)
		assert degree >= 0
		self.degree = degree

		assert isinstance(subspace_dimension, int)
		assert subspace_dimension >= 1
		self.subspace_dimension = subspace_dimension

		if self.degree == 1 and subspace_dimension > 1:
			self.subspace_dimension = 1

		if self.degree == 0:
			self.subspace_dimension = 0

		basis = basis.lower()
		assert basis in ['arnoldi', 'legendre', 'monomial', 'chebyshev', 'laguerre', 'hermite']
		self.basis_name = copy(basis)

		if basis == 'arnoldi':
			self.Basis = ArnoldiPolynomialBasis
		elif basis == 'legendre':
			self.Basis = LegendreTensorBasis
		elif basis == 'monomial':
			self.Basis = MonomialTensorBasis
		elif basis == 'chebyshev':
			self.Basis = ChebyshevTensorBasis
		elif basis == 'laguerre':
			self.Basis = LaguerreTensorBasis
		elif basis == 'hermite':
			self.Basis = HermiteTensorBasis

		assert isinstance(keep_data, bool)
		self.keep_data = keep_data

		assert isinstance(scale, bool)
		self.scale = scale

		assert norm in [1,2,'inf', np.inf], "Invalid norm specified"
		if norm == 'inf': norm = np.inf
		self.norm = norm

		if domain is None:
			self.domain = None
		else:
			assert isinstance(domain, Domain)
			self.domain = deepcopy(domain)


		assert bound in [None, 'lower', 'upper'], "Invalid bound specified"
		self.bound = bound

	def __len__(self):
		return self.U.shape[0]

	def __str__(self):
		return "<PolynomialRidgeApproximation degree %d, subspace dimension %d>" % (self.degree, self.subspace_dimension)


	def fit_fixed_subspace(self, X, fX, U):
		r"""

		"""
		assert U.shape[0] == X.shape[1], "U has %d rows, expected %d based on X" % (U.shape[0], X.shape[1])
		assert U.shape[1] == self.subspace_dimension, "U has %d columns; expected %d" % (U.shape[1], self.subspace_dimension)
		self._finish(X, fX, U)


	def fit(self, X, fX, U0 = None):
		r""" Given samples, fit the polynomial ridge approximation.

		Parameters
		----------
		X : array-like (M, m)
			Input coordinates
		fX : array-like (M,)
			Evaluations of the function at the samples

		"""
		kwargs = self.kwargs

		X = np.array(X)
		fX = np.array(fX).flatten()

		assert X.shape[0] == fX.shape[0], "Dimensions of input do not match"

		# Check if we have enough data to make problem overdetermined
		m = X.shape[1]
		n = self.subspace_dimension
		d = self.degree
		n_param = scipy.special.comb(n+d, d)	# Polynomial contribution
		n_param += m*n - (n*(n+1))//2			# Number of parameters in Grassmann manifold
		if len(fX) < n_param:
			mess = "A polynomial ridge approximation of degree %d and subspace dimension %d of a %d-dimensional function " % (d, n, m)
			mess += "requires at least %d samples to not be underdetermined" % (n_param, )
			raise UnderdeterminedException(mess)

		# Special case where solution is convex and no iteration is required
		if self.subspace_dimension == 1 and self.degree == 1:
			self._U = self._fit_affine(X, fX)
			self.coef = self._fit_coef(X, fX, self._U)
			return


		if U0 is not None:
			# Check that U0 has the right shape
			U0 = np.array(U0)
			assert U0.shape[0] == X.shape[1], "U0 has %d rows, expected %d based on X" % (U0.shape[0], X.shape[1])
			assert U0.shape[1] == self.subspace_dimension, "U0 has %d columns; expected %d" % (U0.shape[1], self.subspace_dimension)
		else:
			U0 = initialize_subspace(X = X, fX = fX)[:,:self.subspace_dimension]

		# Orthogonalize just to make sure the starting value satisfies constraints
		U0 = orth(U0)

		# TODO Implement multiple initializations
		if self.norm == 2 and self.bound == None:
			return self._fit_varpro(X, fX, U0, **kwargs)
		else:
			return self._fit_alternating(X, fX, U0, **kwargs)


	################################################################################
	# Specialized Affine fits
	def _fit_affine(self, X, fX):
		r""" Solves the affine
		"""
		# Normalize the domain
		lb = np.min(X, axis = 0)
		ub = np.max(X, axis = 0)
		dom = BoxDomain(lb, ub)
		XX = np.hstack([dom.normalize(X), np.ones((X.shape[0],1))])

		# Normalize the output
		fX = (fX - np.min(fX))/(np.max(fX) - np.min(fX))

		if self.bound is None:
			if self.norm == 1:
				b = one_norm_fit(XX, fX)
			elif self.norm == 2:
				b = two_norm_fit(XX, fX)
			elif self.norm == np.inf:
				b = inf_norm_fit(XX, fX)
		elif self.bound == 'lower':
			# fX >= XX b
			b = bound_fit(XX, fX, norm = self.norm)
		elif self.bound == 'upper':
			b = bound_fit(-XX, -fX, norm = self.norm)

		U = b[0:-1].reshape(-1,1)
		# Correct for transform
		U = dom._normalize_der().dot(U)
		# Force to have unit norm
		U /= np.linalg.norm(U)
		return U

	def _fit_coef(self, X, fX, U):
		r""" Returns the linear coefficients
		"""
		Y = (U.T @ X.T).T
		self.basis = self.Basis(self.degree, X = Y)
		V = self.basis.V(Y)
		if self.bound is None:
			if self.norm == 1:
				c = one_norm_fit(V, fX)
			elif self.norm == 2:
				c = two_norm_fit(V, fX)
			elif self.norm == np.inf:
				c = inf_norm_fit(V, fX)
			else:
				raise NotImplementedError
		elif self.bound == 'lower':
			c = bound_fit(-V, -fX, norm = self.norm)
		elif self.bound == 'upper':
			c = bound_fit(V, fX, norm = self.norm)

		return c

	def _finish(self, X, fX, U):
		r""" Given final U, rotate and find coefficients
		"""

		Y = (U.T @ X.T).T
		# Step 1: Apply active subspaces to the profile function at samples X
		# to rotate onto the most important directions
		if U.shape[1] > 1 and self.rotate:
			self._U = U
			self.coef = self._fit_coef(X, fX, U)
			grads = self.profile.grad(Y)
			# We only need the short-form SVD
			Ur = scipy.linalg.svd(grads.T, full_matrices = False)[0]
			U = U @ Ur

		self._U = U

		# Step 2: Flip signs such that average slope is positive in the coordinate directions
		if self.rotate:
			self.coef = self._fit_coef(X, fX, U)
			grads = self.profile.grad(Y)
			self._U = U = U.dot(np.diag(np.sign(np.mean(grads, axis = 0))))

		# Step 3: final fit
		self.coef = self._fit_coef(X, fX, U)

	################################################################################
	# VarPro based solution for the 2-norm without bound constraints
	################################################################################

	def _varpro_residual(self, X, fX, U_flat):
		U = U_flat.reshape(X.shape[1],-1)

		#V = self.V(X, U)
		Y = (U.T @ X.T).T
		self.basis = self.Basis(self.degree, Y)
		V = self.basis.V(Y)
		if self.basis_name == 'arnoldi':
			# In this case, V is orthonormal
			c = V.T @ fX
		else:
			c = scipy.linalg.lstsq(V, fX)[0].flatten()
		r = fX - V.dot(c)
		return r

	def _varpro_jacobian(self, X, fX, U_flat):
		# Get dimensions
		M, m = X.shape
		U = U_flat.reshape(X.shape[1],-1)
		m, n = U.shape

		Y = (U.T @ X.T).T
		self.basis = self.Basis(self.degree, Y)
		V = self.basis.V(Y)
		DV = self.basis.DV(Y)

		if isinstance(self.basis, ArnoldiPolynomialBasis):
			# In this case, V is orthonormal
			c = V.T @ fX
			Y = np.copy(V)
			s = np.ones(V.shape[1])
			ZT = np.eye(V.shape[1])
		else:
			c = scipy.linalg.lstsq(V, fX)[0].flatten()
			Y, s, ZT = scipy.linalg.svd(V, full_matrices = False)

		r = fX - V.dot(c)


		N = V.shape[1]
		J1 = np.zeros((M,m,n))
		J2 = np.zeros((N,m,n))

		for ell in range(n):
			for k in range(m):
				DVDU_k = X[:,k,None]*DV[:,:,ell]

				# This is the first term in the VARPRO Jacobian minus the projector out fron
				J1[:, k, ell] = DVDU_k.dot(c)
				# This is the second term in the VARPRO Jacobian before applying V^-
				J2[:, k, ell] = DVDU_k.T.dot(r)

		# Project against the range of V
		J1 -= np.tensordot(Y, np.tensordot(Y.T, J1, (1,0)), (1,0))
		# Apply V^- by the pseudo inverse
		J2 = np.tensordot(np.diag(1./s),np.tensordot(ZT, J2, (1,0)), (1,0))
		J = -( J1 + np.tensordot(Y, J2, (1,0)))
		return J.reshape(J.shape[0], -1)

	def _grassmann_trajectory(self, U_flat, Delta_flat, t):
		Delta = Delta_flat.reshape(-1, self.subspace_dimension)
		U = U_flat.reshape(-1, self.subspace_dimension)
		Y, s, ZT = scipy.linalg.svd(Delta, full_matrices = False, lapack_driver = 'gesvd')
		UZ = np.dot(U, ZT.T)
		U_new = np.dot(UZ, np.diag(np.cos(s*t))) + np.dot(Y, np.diag(np.sin(s*t)))
		U_new = orth(U_new).flatten()
		return U_new

	def _fit_varpro(self, X, fX, U0, **kwargs):


		def gn_solver(J_flat, r):
			Y, s, ZT = scipy.linalg.svd(J_flat, full_matrices = False, lapack_driver = 'gesvd')
			# Apply the pseudoinverse
			n = self.subspace_dimension
			Delta_flat = -ZT[:-n**2,:].T.dot(np.diag(1/s[:-n**2]).dot(Y[:,:-n**2].T.dot(r)))
			return Delta_flat, s[:-n**2]

		def jacobian(U_flat):
			return self._varpro_jacobian(X, fX, U_flat)

		def residual(U_flat):
			return self._varpro_residual(X, fX, U_flat)

		U0_flat = U0.flatten()
		U_flat, info = gauss_newton(residual, jacobian, U0_flat,
			trajectory = self._grassmann_trajectory, gnsolver = gn_solver, **kwargs)

		U = U_flat.reshape(-1, self.subspace_dimension)

		self._finish(X, fX, U)

	################################################################################
	# Generic residual and Jacobian
	################################################################################

	def _residual(self, X, fX, U_c):
		M, m = X.shape
		N = len(self.basis)
		n = self.subspace_dimension

		# Extract U and c
		U = U_c[:m*n].reshape(m,n)
		c = U_c[m*n:].reshape(N)

		# Construct basis
		#V = self.V(X, U)
		Y = (U.T @ X.T).T
		V = self.basis.V(Y)
		res = V @ c - fX
		return res

	def _jacobian(self, X, fX, U_c):
		M, m = X.shape
		N = len(self.basis)
		n = self.subspace_dimension

		# Extract U and c
		U = U_c[:m*n].reshape(m,n)
		c = U_c[m*n:].reshape(N)

		# Re-initialize basis
		Y = (U.T @ X.T).T
		self.basis = self.Basis(self.degree, Y)
		V = self.basis.V(Y)

		# Derivative of V with respect to U with c fixed
		DVDUc = np.zeros((M,m,n))
		#DV = self.DV(X, U) 	# Size (M, N, n)
		DV = self.basis.DV(Y)
		for k in range(m):
			for ell in range(n):
				DVDUc[:,k,ell] = X[:,k]*np.dot(DV[:,:,ell], c)

		# Derivative with respect to linear component
		#V = self.V(X, U)

		# Total Jacobian
		jac = np.hstack([DVDUc.reshape(M,-1), V])
		return jac


	def _trajectory(self, X, fX, U_c, pU_pc, alpha):
		r""" For the trajectory through the sup-norm space, we automatically compute optimal c
		and advance U along the geodesic

		"""
		M, m = X.shape
		N = len(self.basis)
		n = self.subspace_dimension

		# Split components
		U = orth(U_c[:m*n].reshape(m,n))
		c = U_c[m*n:].reshape(N)

		Delta = pU_pc[:m*n].reshape(m,n)
		pc = pU_pc[m*n:].reshape(N)

		# Orthogonalize
		Delta = Delta - U.dot(U.T.dot(Delta))

		# Compute the step along the Geodesic
		Y, s, ZT = scipy.linalg.svd(Delta, full_matrices = False, lapack_driver = 'gesvd')
		U_new = np.dot(np.dot(U,ZT.T), np.diag(np.cos(s*alpha))) + np.dot(Y, np.diag(np.sin(s*alpha)))

		# TODO: align U and U_new to minimize Frobenius norm error
		# right the small step termination criteria is never triggering because U_new and U have different orientations

		# Solve a convex problem to actually compute optimal c
		c = self._fit_coef(X, fX, U_new)

		return np.hstack([U_new.flatten(), c.flatten()])

	def _fit_alternating(self, X, fX, U0, **kwargs):
		M, m = X.shape
		n = self.subspace_dimension

		def residual(U_c):
			r = self._residual(X, fX, U_c)
			return r

		def jacobian(U_c):
			m = X.shape[1]
			n = self.subspace_dimension
			U = U_c[:m*n].reshape(m,n)
			#self.set_scale(X, U)
			J = self._jacobian(X, fX, U_c)
			return J

		# Trajectory
		trajectory = lambda U_c, p, alpha: self._trajectory(X, fX, U_c, p, alpha)

		# Initialize parameter values
		#self.set_scale(X, U0)
		c0 = self._fit_coef(X, fX, U0)
		U_c0 = np.hstack([U0.flatten(), c0])

		# Add orthogonality constraints to search direction
		# Recall pU.T @ U == 0 is a requirement for Grassmann optimization
		def search_constraints(U_c, pU_pc):
			M, m = X.shape
			N = len(self.basis)
			n = self.subspace_dimension
			U = U_c[:m*n].reshape(m,n)
			constraints = [ pU_pc[k*m:(k+1)*m].__rmatmul__(U.T) == np.zeros(n) for k in range(n)]
			return constraints

		# setup lower/upper bound into SLP solver
		obj_lb = None
		obj_ub = None
		if self.bound == 'lower':
			obj_ub = np.zeros(fX.shape)
		elif self.bound == 'upper':
			obj_lb = np.zeros(fX.shape)

		# Perform optimization
		U_c = sequential_lp(residual, U_c0, jacobian, trajectory = trajectory,
			obj_lb = obj_lb, obj_ub = obj_ub,
			search_constraints = search_constraints, norm = self.norm, **kwargs)

		# Store solution
		U = U_c[:m*n].reshape(m,n)
		self._finish(X, fX, U)



import numpy as np
import pandas as pd
from scipy.stats import norm, beta, lognorm, weibull_min
import unittest

from context import grama as gr
from context import models, data

## Marginals function tests
##################################################
class TestMarginalTools(unittest.TestCase):
    def setUp(self):
        self.median = np.median(data.df_stang.E)

    def test_gkde(self):
        mg_gkde = gr.marg_gkde(data.df_stang.E)

        d_gkde = mg_gkde.d(np.array([1, 10000, 10400, 10800, 1e6]))
        p_gkde = mg_gkde.p(np.array([1, 10000, 10400, 10800, 1e6]))
        q_gkde = mg_gkde.q(np.array([0.0, 0.25, 0.50, 0.75, 1.0]))
        mg_gkde.summary()

        self.assertTrue(np.isclose(q_gkde[2], self.median, atol=0, rtol=0.05))

        # Test random variable sample
        n = 10
        n_gkde = mg_gkde.r(n)
        self.assertTrue(len(n_gkde) == n)

        # Copy
        mg_new = mg_gkde.copy()
        self.assertTrue(all(np.isclose(
            mg_new.d(q_gkde),
            mg_gkde.d(q_gkde),
        )))
        self.assertTrue(all(np.isclose(
            mg_new.p(q_gkde),
            mg_gkde.p(q_gkde),
        )))

    def test_fit(self):
        mg_norm = gr.marg_fit("norm", data.df_stang.E)

        d_norm = mg_norm.d(np.array([10000, 10400, 10800]))
        p_norm = mg_norm.p(np.array([10000, 10400, 10800]))
        q_norm = mg_norm.q(np.array([0.25, 0.50, 0.75]))
        mg_norm.summary()

        self.assertTrue(np.isclose(q_norm[1], self.median, atol=0, rtol=0.05))

        # Test random variable sample
        n = 10
        n_norm = mg_norm.r(n)
        self.assertTrue(len(n_norm) == n)

        ## Raises error when dataframe passed
        with self.assertRaises(ValueError):
            gr.marg_fit("norm", data.df_stang)
        with self.assertRaises(ValueError):
            gr.marg_gkde(data.df_stang)

    def test_mom(self):
        ## Test for accuracy
        mg_mom = gr.marg_mom(
            "beta",
            mean=1,
            sd=1,
            skew=0,
            kurt=2,
        )

        self.assertTrue(all(np.isclose(
            beta(**mg_mom.d_param).stats("mvsk"),
            np.array([1, 1, 0, 2 - 3])
        )))

        ## Test COV specification
        mg_cov = gr.marg_mom(
            "beta",
            mean=1,
            cov=2,
            skew=0,
            kurt=2,
        )

        self.assertTrue(all(np.isclose(
            beta(**mg_cov.d_param).stats("mvsk"),
            np.array([1, 4, 0, 2 - 3])
        )))

        ## Test 2-parameter lognormal
        mg_log = gr.marg_mom(
            "lognorm",
            mean=5e4,
            var=5e2**2,
            floc=0,
        )

        self.assertTrue(all(np.isclose(
            lognorm(**mg_log.d_param).stats("mv"),
            np.array([5e4, 5e2**2])
        )))

        ## Test 2-parameter Weibull fit
        mg_weibull2 = gr.marg_mom(
            "weibull_min",
            mean=5e4,
            var=5e2**2,
            floc=0,
        )

        self.assertTrue(all(np.isclose(
            weibull_min(**mg_weibull2.d_param).stats("mv"),
            np.array([5e4, 5e2**2])
        )))

        ## Test bounded uniform
        mg_unif_bnd = gr.marg_mom("uniform", lo=-1, up=+1)
        self.assertTrue(all(np.isclose(
            (mg_unif_bnd.d_param["loc"], mg_unif_bnd.d_param["scale"]),
            (-1, +2)
        )))

        ## Test bounded beta
        mg_beta_bnd1 = gr.marg_mom("beta", lo=-1, up=+1, mean=0, var=1/8)
        self.assertTrue(all(np.isclose(
            beta(**mg_beta_bnd1.d_param).stats("mv"),
            np.array([0, 1/8])
        )))
        mg_beta_bnd2 = gr.marg_mom("beta", lo=-1, up=+1, mean=0.2, var=1/8)
        self.assertTrue(all(np.isclose(
            beta(**mg_beta_bnd2.d_param).stats("mv"),
            np.array([0.2, 1/8])
        )))

        ## Test invariants
        # Must provide mean
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", sd=1)
        # Must provide sd or var
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", mean=1)
        # Must provide sufficient parameters
        with self.assertRaises(ValueError):
            gr.marg_mom("lognorm", mean=1, sd=1)
        # Must not overdetermine
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", mean=1, sd=1, skew=0, kurt=3)
        # For beta; skew == 0 and kurt == 4 is infeasible
        with self.assertRaises(RuntimeError):
            gr.marg_mom("beta", mean=1, sd=1, skew=0, kurt=4)
        # Cannot provide more than one of `sd`, `cov`, or `var`
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", mean=1, sd=1, var=1)
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", mean=1, sd=1, cov=1)
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", mean=1, cov=1, var=1)
        with self.assertRaises(ValueError):
            gr.marg_mom("norm", mean=1, sd=1, var=1, cov=1)
        # Cannot provide both kurt and kurt_excess
        with self.assertRaises(ValueError):
            gr.marg_mom("lognorm", mean=1, sd=1, kurt=1, kurt_excess=-2)
        # Cannot use lo or up with non-bounded dist
        with self.assertRaises(ValueError):
            gr.marg_mom("lognorm", mean=1, sd=1, kurt=1, kurt_excess=-2, lo=0)
        with self.assertRaises(ValueError):
            gr.marg_mom("lognorm", mean=1, sd=1, kurt=1, kurt_excess=-2, up=0)

    def test_trunc(self):
        mg_base = gr.marg_mom("norm", mean=0, sd=1)

        ## Two-sided truncation
        mg_trunc1 = gr.marg_trunc(mg_base, lo=-2, up=+2)
        # Summary
        mg_trunc1.summary()

        # Quantiles
        self.assertTrue(abs(mg_trunc1.q(0) + 2) < 1e-6)
        self.assertTrue(abs(mg_trunc1.q(1) - 2) < 1e-6)
        # PDF truncated
        self.assertTrue(abs(mg_trunc1.d(-2.1)) < 1e-6)
        self.assertTrue(abs(mg_trunc1.d(+2.1)) < 1e-6)
        # CDF truncated
        self.assertTrue(abs(mg_trunc1.p(-2.1) - 0) < 1e-6)
        self.assertTrue(abs(mg_trunc1.p(+2.1) - 1) < 1e-6)

        # Vector evaluation
        qv = mg_trunc1.q(np.array([0, 0.5, 1]))
        self.assertTrue(all(np.isclose(qv, np.array([-2, 0, +2]))))
        dv = mg_trunc1.d(np.array([-2.1, 0, +2.1]))
        self.assertTrue(all(np.isclose(
            dv,
            np.array([0, mg_base.d(0) / (mg_base.p(2) - mg_base.p(-2)), 0])
        )))
        pv = mg_trunc1.p(np.array([-2.1, 0, +2.1]))
        self.assertTrue(all(np.isclose(pv, np.array([0, 0.5, 1]))))

        ## Lower truncation
        mg_trunc_lo = gr.marg_trunc(mg_base, lo=0)
        # Quantiles
        self.assertTrue(abs(mg_trunc_lo.q(0) - 0) < 1e-6)
        self.assertTrue(np.isinf(mg_trunc_lo.q(1)))
        # PDF truncated
        self.assertTrue(abs(mg_trunc_lo.d(-0.1)) < 1e-6)
        # PDF doubled for truncation at center
        self.assertTrue(abs(mg_trunc_lo.d(0) - mg_base.d(0) * 2) < 1e-6)
        # CDF truncated
        self.assertTrue(abs(mg_trunc_lo.p(-0.1) - 0) < 1e-6)
        self.assertTrue(abs(mg_trunc_lo.p(+np.Inf) - 1) < 1e-6)

        ## Upper truncation
        mg_trunc_up = gr.marg_trunc(mg_base, up=0)
        # Quantiles
        self.assertTrue(np.isinf(mg_trunc_up.q(0)))
        self.assertTrue(abs(mg_trunc_up.q(1) - 0) < 1e-6)
        # PDF truncated
        self.assertTrue(abs(mg_trunc_up.d(+0.1)) < 1e-6)
        # PDF doubled for truncation at center
        self.assertTrue(abs(mg_trunc_up.d(0) - mg_base.d(0) * 2) < 1e-6)
        # CDF truncated
        self.assertTrue(abs(mg_trunc_up.p(-np.Inf) - 0) < 1e-6)
        self.assertTrue(abs(mg_trunc_up.p(+0.1) - 1) < 1e-6)

        # Copy
        mg_copy = mg_trunc1.copy()
        self.assertTrue(all(np.isclose(
            mg_copy.q(np.array([0, 0.5, 1])),
            mg_trunc1.q(np.array([0, 0.5, 1])),
        )))

        # Invalid truncation
        with self.assertRaises(ValueError):
            gr.marg_trunc(mg_base)

# --------------------------------------------------

class TestMarginal(unittest.TestCase):
    def setUp(self):
        self.marginal_named = gr.MarginalNamed(
            d_name="norm", d_param={"loc": 0, "scale": 1}
        )

    def test_fcn(self):

        ## Invoke summary
        self.marginal_named.summary()

        ## Correct values for normal distribution
        self.assertTrue(self.marginal_named.d(0.5) == norm.pdf(0.5))
        self.assertTrue(self.marginal_named.p(0.5) == norm.cdf(0.5))
        self.assertTrue(self.marginal_named.q(0.5) == norm.ppf(0.5))



import re
import unittest

from context import grama as gr
from context import data
from numpy import NaN, random
from pandas import DataFrame, RangeIndex
from pandas.testing import assert_frame_equal


class TestPivotLonger(unittest.TestCase):
    """Test implementation of pivot_longer
    """
    def test_pivot_longer(self):
        """ Test basic functionality of pivot_longer
        """
        wide = gr.df_make(One=[1,2,3], Two=[4,5,6])
        long = gr.tran_pivot_longer(
            wide,
            columns=("One","Two"),
            names_to="columns",
            values_to="values"
        )
        expected = gr.df_make(
            columns=["One","One","One","Two","Two","Two"],
            values=[1,2,3,4,5,6]
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_index_not_rangeindex(self):
        """ Test if pivot_longer makes a RangeIndex if current index is not
            a RangeIndex and preserves the orignal as new column named "index"
        """
        wide = DataFrame(
            {
                "One": {"A": 1.0, "B": 2.0, "C": 3.0},
                "Two": {"A": 1.0, "B": 2.0, "C": 3.0},
            }
        )
        long = gr.tran_pivot_longer(
            wide,
            columns=("One","Two"),
            names_to="columns",
            values_to="values"
        )
        expected = DataFrame(
            {
                "index": ["A", "B", "C", "A", "B", "C"],
                "columns": ["One", "One", "One", "Two", "Two", "Two"],
                "values": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
            }
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_rename_not_rangeindex(self):
        """ Test if pivot_longer makes a RangeIndex if current index is not
            a RangeIndex and preserves the orignal as a new column "index_to"
        """
        wide = DataFrame(
            {
                "One": {"A": 1.0, "B": 2.0, "C": 3.0},
                "Two": {"A": 1.0, "B": 2.0, "C": 3.0},
            }
        )
        long = gr.tran_pivot_longer(
            wide,
            columns=("One","Two"),
            index_to="idx",
            names_to="columns",
            values_to="values"
        )
        expected = DataFrame(
            {
                "idx": ["A", "B", "C", "A", "B", "C"],
                "columns": ["One", "One", "One", "Two", "Two", "Two"],
                "values": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
            }
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_representation_index(self):
        """ Test if pivot_longer produces a representation index for nx2
            DataFrame that has an index_to call
        """
        wide = gr.df_make(A=[1,2,3], B=[4,5,6])
        long = gr.tran_pivot_longer(
            wide,
            columns=("A", "B"),
            index_to="index",
            names_to="var",
            values_to="value"
        )
        expected = DataFrame(
            {
                "index": [0,1,2,0,1,2],
                "var": ["A","A","A","B","B","B"],
                "value": [1,2,3,4,5,6]
            }
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_no_representation_index(self):
        """ Test if pivot_longer does not produce a representation index for nx2
            DataFrame that has no index_to call
        """
        wide = gr.df_make(A=[1,2,3], B=[4,5,6])
        long = gr.tran_pivot_longer(
            wide,
            columns=("A", "B"),
            names_to="var",
            values_to="value"
        )
        expected = DataFrame(
            {
                "var": ["A","A","A","B","B","B"],
                "value": [1,2,3,4,5,6]
            }
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_rep_index_groups(self):
        """ Test if pivot_longer adds a representation index correctly when
            there is multiple groupings of observations
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            index_to = "idx",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to="var",
            values_to="val"
        )

        long_index = list(long["idx"])
        mod = max(long_index)
        single_index = list(range(0, mod))

        result = False
        if single_index == long_index[0:mod]:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_select(self):
        """ Test if pivot_longer is compatible with gr.tran_select or any input
            of a DataFrame instead as 'columns' argument
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            columns=(gr.tran_select(stang,gr.matches("\\d+"))),
            names_to="var",
            values_to="val"
        )
        expected = gr.tran_pivot_longer(
            stang,
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to="var",
            values_to="val"
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_matches(self):
        """ Test if pivot_longer is compatible with gr.matches as columns input
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            columns = gr.matches("\\d+"),
            names_to="var",
            values_to="val"
        )

        expected = gr.tran_pivot_longer(
            stang,
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to="var",
            values_to="val"
        )

        assert_frame_equal(long, expected)


    def test_pivot_longer_matches_error(self):
        """ Test if selection helper retrieves no matches
        """
        with self.assertRaises(ValueError):
            stang = data.df_stang_wide
            long = gr.tran_pivot_longer(
                stang,
                columns = gr.matches("0123"),
                names_to="var",
                values_to="val"
            )


    def test_pivot_longer_names_pattern(self):
        """ Test if pivot_longer properly works with names_pattern
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            names_pattern="(E|mu)_(\\d+)",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=("property", "angle"),
            values_to="val"
        )

        names_to = ["property","angle"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_and_pattern(self):
        """ Test if pivot_longer raises a ValueError if both names_sep and
            names_pattern are called
        """
        with self.assertRaises(ValueError):
            stang = data.df_stang_wide
            long = gr.tran_pivot_longer(
                stang,
                names_pattern="(E|mu)_(\\d+)",
                names_sep="_",
                columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
                names_to=("property", "angle"),
                values_to="val"
            )


    def test_pivot_longer_names_sep(self):
        """ Test if pivot_longer properly works with names_sep argument
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            names_sep="_",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=("property", "angle"),
            values_to="val"
        )

        names_to = ["property","angle"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_regex(self):
        """ Test if pivot_longer properly works with names_sep argument being
            regex expression
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            names_sep=r'\_',
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=("property", "angle"),
            values_to="val"
        )
        names_to = ["property","angle"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_position(self):
        """ Test if pivot_longer works with names_sep argument being a position
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            names_sep=[-3],
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=("property", "angle"),
            values_to="val"
        )
        names_to = ["property","angle"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_thrice(self):
        """ Test if pivot_longer properly works with names_sep having to split
            columns into 3 or more
        """
        wide = gr.df_make(A_1_hello=[1,2,3], B_2_bye=[4,5,6])
        long = gr.tran_pivot_longer(
            wide,
            names_sep="_",
            columns=["A_1_hello","B_2_bye"],
            names_to=("letter","num","saying"),
            values_to="val"
        )
        names_to = ["letter","num","saying"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_position_thrice(self):
        """ Test if pivot_longer works with names_sep argument being a position
        """
        wide = gr.df_make(A_1_hello=[1,2,3], B_2_bye=[4,5,6])
        long = gr.tran_pivot_longer(
            wide,
            names_sep=[1, 3],
            columns=["A_1_hello","B_2_bye"],
            names_to=("letter","num","saying"),
            values_to="val"
        )
        names_to = ["letter","num","saying"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_multiple_seps(self):
        """ Test if pivot_longer properly works with names_sep having column
            names with varying amount of seps
        """
        wide = gr.df_make(A_1_hello=[1,2,3], B_2=[4,5,6])
        long = gr.tran_pivot_longer(
            wide,
            names_sep="_",
            columns=["A_1_hello","B_2"],
            names_to=("letter","num","saying"),
            values_to="val"
        )
        names_to = ["letter","num","saying"]
        names_to_check = [x for x in long.columns.values if x in names_to]

        result = False
        if names_to == names_to_check:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_and_index_to(self):
        """ Test if pivot_longer works with names_sep and index_to arguments
            together
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            index_to="idx",
            names_sep="_",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=("property", "angle"),
            values_to="val"
        )

        long_index = list(long["idx"])
        mod = max(long_index)
        single_index = list(range(0, mod))

        check = ["property","angle"]
        col_check = [x for x in long.columns.values if x in check]

        result = False
        if single_index == long_index[0:mod]:
            if check == col_check:
                result = True

        self.assertTrue(result)


    def test_pivot_longer_names_sep_names_to_is_1(self):
        """ Test if pivot_longer raises an error for only giving one value to
            names_to and calling names_sep
        """
        with self.assertRaises(TypeError):
            stang = data.df_stang_wide
            long = gr.tran_pivot_longer(
                stang,
                names_sep="_",
                columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
                names_to="property",
                values_to="val"
            )


    def test_pivot_longer_dot_value(self):
        """ Test pivot_longer when it receives the .value input for names_to
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=".value",
            values_to="val"
        )
        check = ["E_00","mu_00","E_45","mu_45","E_90","mu_90"]
        col_check = [x for x in long.columns.values if x in check]

        result = False
        if set(col_check) == set(check):
            result = True

        self.assertTrue(result)

    def test_pivot_longer_dot_value_and_names_sep(self):
        """ Test pivot_longer when it receives the .value and names_sep
        """
        DF = gr.Intention()
        wide = gr.df_make(x=range(0, 6))
        wide = gr.tran_mutate(
            wide,
            y_Trend=DF.x**2,
            y_Variability=random.normal(size=6),
            y_Mixed=DF.x**2 + random.normal(size=6),
        )

        long = gr.tran_pivot_longer(
            wide,
            columns=["y_Trend", "y_Variability", "y_Mixed"],
            names_to=(".value", "type"),
            names_sep="_"
        )

        check = ["x", "type", "y"]
        col_check = [x for x in long.columns.values if x in check]

        result = False
        if set(col_check) == set(check):
            result = True

        self.assertTrue(result)


    def test_pivot_longer_dot_value_and_index_to(self):
        """ Test pivot_longer when it receives the .value input and an input
            for 'index_to'
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            index_to="idx",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=".value",
            values_to="val"
        )

        check = ["E_00","mu_00","E_45","mu_45","E_90","mu_90"]
        col_check = [x for x in long.columns.values if x in check]

        result = False
        if set(col_check) == set(check):
            if long.columns.values[0] == "idx":
                result = True

        self.assertTrue(result)


    def test_pivot_longer_dot_value_and_name(self):
        """ Test pivot_longer when it receives the .value and another input for
            names_to
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            names_sep="_",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to=(".value", "angle"),
            values_to="val"
        )
        columns = list(long.columns.values)
        E_count = columns.count("E")
        mu_count = columns.count("mu")

        result = False
        if E_count > 0 and mu_count > 0:
            result = True

        self.assertTrue(result)


    def test_pivot_longer_dot_value_in_the_middle(self):
        """ Test pivot_longer when it receives the .value input in between at
            least 2 other inputs in 'names_to'
        """
        stang = data.df_stang_wide
        # rename columns for this test
        for i in stang.columns.values:
            stang.rename(columns={i: ('NA_'+i)},inplace=True)
        stang.rename(columns={'NA_thick': 'thick'},inplace=True)
        stang.rename(columns={'NA_alloy': 'alloy'},inplace=True)

        long = gr.tran_pivot_longer(
            stang,
            names_sep="_",
            columns=["NA_E_00","NA_mu_00","NA_E_45","NA_mu_45","NA_E_90","NA_mu_90"],
            names_to=("null", ".value", "angle"),
            values_to="val"
        )

        columns = list(long.columns.values)
        null_count = columns.count("null")
        E_count = columns.count("E")
        mu_count = columns.count("mu")

        result = False
        if E_count > 0 and mu_count > 0 and null_count > 0:
            result = True

        self.assertTrue(result)

        # undo column renaming for future tests
        null = 'NA_'
        for i in stang.columns.values:
            if null in i:
                stang.rename(columns={i: i.lstrip(null)},inplace=True)


class TestPivotWider(unittest.TestCase):
    """Test implementation of pivot_wider
    """
    def test_pivot_wider(self):
        long = DataFrame(
            {
                "loo": ["A", "B", "C", "A", "B", "C"],
                "columns": ["One", "One", "One", "Two", "Two", "Two"],
                "values": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
            }
        )
        wide = gr.tran_pivot_wider(
            long,
            indexes_from="loo",
            names_from="columns",
            values_from="values"
        )
        expected = DataFrame(
            {
                "One": {"A": 1.0, "B": 2.0, "C": 3.0},
                "Two": {"A": 1.0, "B": 2.0, "C": 3.0},
            }
        )
        expected.insert(0,"loo",["A","B","C"])
        expected.index=RangeIndex(start=0,stop=len(expected),step=1)

        assert_frame_equal(wide, expected)


    def test_pivot_wider_duplicate_entries(self):
        """ Test if pivot_wider raise an error for duplicate entries
        """
        with self.assertRaises(ValueError):
            stang = data.df_stang_wide
            long = gr.tran_pivot_longer(
                stang,
                columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
                names_to="var",
                values_to="val"
            )
            wide = gr.tran_pivot_wider(
                long,
                names_from="var",
                values_from="val"
            )


    def test_pivot_wider_representation_index(self):
        """ Test if pivot_wider can handle duplicate entries if a representation
            index is present
        """
        stang = data.df_stang_wide
        long = gr.tran_pivot_longer(
            stang,
            index_to = "idx",
            columns=["E_00","mu_00","E_45","mu_45","E_90","mu_90"],
            names_to="var",
            values_to="val"
        )
        wide = gr.tran_pivot_wider(
            long,
            names_from="var",
            values_from="val"
        )
        expected = gr.df_make(
            idx=[0,1,2,3,4,5,6,7,8],
            thick=[0.022,0.022,0.032,0.032,0.064,0.064,0.081,0.081,0.081],
            alloy=[" al_24st"," al_24st"," al_24st"," al_24st"," al_24st",
                " al_24st"," al_24st"," al_24st"," al_24st"],
            E_00=[10600.000,10600.000,10400.000,10300.000,10500.000,10700.000,
                10000.000,10100.000,10000.000],
            E_45=[10700.000,10500.000,10400.000,10500.000,10400.000,10500.00,
                10000.000,9900.000,-1.0],
            E_90=[10500.000,10700.000,10300.000,10400.000,10400.000,10500.000,
                9900.000,10000.000,9900.000],
            mu_00=[0.321,0.323,0.329,0.319,0.323,0.328,0.315,0.312,0.311],
            mu_45=[0.329,0.331,0.318,0.326,0.331,0.328,0.320,0.312,-1.000],
            mu_90=[0.310,0.323,0.322,0.330,0.327,0.320,0.314,0.316,0.314]
        )

        assert_frame_equal(wide, expected)


    def test_pivot_wider_representation_index_multiple_cols(self):
        long = DataFrame(
            {
                "loo": ["A", "B", "C", "A", "B", "C"],
                "foo": ["X", "Y", "Z", "X", "Y", "Z"],
                "columns": ["One", "One", "One", "Two", "Two", "Two"],
                "values": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
            }
        )
        wide = gr.tran_pivot_wider(
            long,
            indexes_from=("loo","foo"),
            names_from="columns",
            values_from="values"
        )
        expected = DataFrame(
            {
                "One": {"A": 1.0, "B": 2.0, "C": 3.0},
                "Two": {"A": 1.0, "B": 2.0, "C": 3.0},
            }
        )
        expected.insert(0,"loo",["A","B","C"])
        expected.insert(1,"foo",["X","Y","Z"])
        expected.index=RangeIndex(start=0,stop=len(expected),step=1)

        assert_frame_equal(wide, expected)


    def test_pivot_wider_NaN_entries(self):
        """ Test if pivot_wider returns a table with NaN values for unspecified
            entries that have no represenational index
        """
        original = gr.df_make(A=[1,2,3], B=[4,5,6])
        long = gr.tran_pivot_longer(
            original,
            columns=("A", "B"),
            names_to="var",
            values_to="value"
        )
        wide = gr.tran_pivot_wider(
            long,
            names_from="var",
            values_from="value"
        )
        expected = gr.df_make(
            A=[1,2,3,NaN,NaN,NaN],
            B=[NaN,NaN,NaN,4,5,6]
        )

        assert_frame_equal(wide, expected)



import unittest

from context import grama as gr
from numpy import isnan
from pandas import Series, Index

##==============================================================================
## string helper tests
##==============================================================================
class TestStringHelpers(unittest.TestCase):
    def setUp(self):
        self.s_true = "foo"
        self.s_false = "bar"
        self.s_ind1 = "afoo"

        self.S_true = ["foo"]
        self.S_false = ["bar"]
        self.S_ind1 = ["afoo"]

    def test_str_c(self):
        ## Correct errors
        with self.assertRaises(ValueError):
            gr.str_c([1, 2], [1, 2, 3])

        ## Correct behavior
        S0_true = Series(["foo"])
        S0_comp = gr.str_c("foo")
        self.assertTrue(S0_true.equals(S0_comp))

        S1_true = Series(["x0", "x1"])
        S1_comp = gr.str_c("x", [0, 1])
        self.assertTrue(S1_true.equals(S1_comp))

        S2_true = Series(["x0", "x1"])
        S2_comp = gr.str_c("x", [0, 1])
        self.assertTrue(S2_true.equals(S2_comp))

        # Catch series index issues
        S3_true = Series(["x0", "x1"])
        S_tmp = Series([0, 1])
        S_tmp.index = Index(["a", "b"])

        S3_comp = gr.str_c("x", S_tmp)
        self.assertTrue(S3_true.equals(S3_comp))

        # Catch length-one issue
        S4_true = Series(["0x"])
        S4_comp = gr.str_c(Series([0], name="test"), "x")
        self.assertTrue(S4_true.equals(S4_comp))

    def test_str_detect(self):
        self.assertTrue(gr.str_detect(self.s_true, "foo"))
        self.assertTrue(not gr.str_detect(self.s_false, "foo"))
        self.assertTrue(all(gr.str_detect(self.S_true, "foo")))
        self.assertTrue(not any(gr.str_detect(self.S_false, "foo")))

    def test_str_locate(self):
        R_ind0 = gr.str_locate(self.S_true, "foo")
        R_empty = gr.str_locate(self.S_false, "foo")

        self.assertTrue(gr.str_locate(self.s_true, "foo") == [0])
        self.assertTrue(len(gr.str_locate(self.s_false, "foo")) == 0)
        self.assertTrue(R_ind0[0] == [0])
        self.assertTrue(len(R_empty[0]) == 0)

    def test_str_which(self):
        R_ind1 = gr.str_which(self.S_ind1, "foo")
        R_nan = gr.str_which(self.S_false, "foo")

        self.assertTrue(gr.str_which(self.s_ind1, "foo") == 1)
        self.assertTrue(isnan(gr.str_which(self.s_false, "foo")))
        self.assertTrue(R_ind1[0] == 1)
        self.assertTrue(isnan(R_nan[0]))

    def test_str_count(self):
        s_c2 = "foofoo"
        s_c0 = "bar"

        S_c2 = ["foofoo"]
        S_c0 = ["bar"]

        R_c2 = gr.str_count(S_c2, "foo")
        R_c0 = gr.str_count(S_c0, "foo")

        self.assertTrue(gr.str_count(s_c2, "foo") == 2)
        self.assertTrue(gr.str_count(s_c0, "foo") == 0)
        self.assertTrue(R_c2[0] == 2)
        self.assertTrue(R_c0[0] == 0)

    def test_str_replace(self):
        self.assertTrue(gr.str_replace("foofoo", "foo", "barbar") == "barbarfoo")
        self.assertTrue(gr.str_replace("barbar", "foo", "bar") == "barbar")

        R = gr.str_replace(["foo", "nope"], "foo", "bar")
        self.assertTrue(R[0] == "bar")
        self.assertTrue(R[1] == "nope")

    def test_str_replace_all(self):
        self.assertTrue(gr.str_replace_all("foofoo", "foo", "bar") == "barbar")

        R = gr.str_replace_all(["foofoo", "nope"], "foo", "bar")
        self.assertTrue(R[0] == "barbar")
        self.assertTrue(R[1] == "nope")

    def test_str_split(self):
        self.assertTrue(gr.str_split("x-y", "-") == ["x", "y"])

        S0 = gr.str_split(["x-y", "u-v-w"], "-")
        self.assertTrue(S0[0] == ["x", "y"])
        self.assertTrue(S0[1] == ["u", "v", "w"])

        S1 = gr.str_split(["x-y", "u-v-w"], "-", maxsplit=1)
        self.assertTrue(S1[0] == ["x", "y"])
        self.assertTrue(S1[1] == ["u", "v-w"])

    def test_str_sub(self):
        s_base0 = "foofoo"
        s_sub0 = "foo"
        s_base1 = "f"
        s_sub1 = "f"

        S_base0 = ["foofoo", "bar"]
        S_sub0 = ["foo", "bar"]

        self.assertTrue(gr.str_sub(s_base0, end=3), s_sub0)
        self.assertTrue(gr.str_sub(s_base1, end=3), s_sub1)
        self.assertTrue(all(gr.str_sub(S_base0, end=3) == S_sub0))

    def test_str_extract(self):
        s_base0 = "foo_bar"
        s_ext0 = "_bar"
        s_base1 = "foo123"
        s_ext1 = "123"

        S_base0 = ["foo_bar", "foo_biddy"]
        S_ext0 = ["_bar", "_biddy"]

        self.assertTrue(gr.str_extract(s_base0, "_\\w+"), s_ext0)
        self.assertTrue(gr.str_extract(s_base1, "\\d+"), s_ext1)

        res = gr.str_extract(S_base0, "_\\w+")
        print(res)  ## DEBUG
        self.assertTrue(all(res == S_ext0))
